# Generated by Django 2.2.5 on 2019-09-29 21:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0008_auto_20190817_1815'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assignedprotection',
            name='state',
            field=models.CharField(choices=[('a', 'disabled'), ('b', 'enabled'), ('c', 'instant_fail')], default='b', help_text='State of the protection.', max_length=1),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-04 17:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0011_auto_20191230_1305'),
        ('spider_tags', '0007_taglayout_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='spidertag',
            name='associated',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent'),
        ),
    ]



# Generated by Django 3.0.1 on 2019-12-30 13:51

from django.db import migrations


def move_to_datacontent(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    WebConfig = apps.get_model("spider_webcfg", "WebConfig")
    DataContent = apps.get_model("spider_base", "DataContent")

    for assigned in AssignedContent.objects.filter(
        ctype__code=WebConfig._meta.model_name
    ):
        d = DataContent(associated=assigned)
        content = WebConfig.objects.get(id=assigned.object_id)
        d.save()
        assigned.attachedblobs.create(
            unique=True, name="config", blob=content.config
        )


class Migration(migrations.Migration):

    dependencies = [
        ('spider_webcfg', '0003_auto_20190708_1341'),
        ('spider_base', '0011_auto_20191230_1305'),
    ]
    run_before = [
        ('spider_base', '0012_auto_20191230_1305'),
    ]

    operations = [
        migrations.RunPython(move_to_datacontent),
    ]



"""spkcspider URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from django.apps import apps
from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.sitemaps import views as sitemaps_views
from django.urls import include, path
from django.views.decorators.cache import cache_page
from django.views.generic.base import RedirectView
from spkcspider.apps.spider.functions import admin_login
from spkcspider.apps.spider.sitemaps import sitemaps
from spkcspider.apps.spider.views import ComponentPublicIndex
from spkcspider.utils.settings import get_settings_func

favicon_view = RedirectView.as_view(
    url='{}spider_base/favicon.svg'.format(settings.STATIC_URL), permanent=True
)
robots_view = RedirectView.as_view(
    url='{}spider_base/robots.txt'.format(settings.STATIC_URL), permanent=True
)

# disable admin login page
admin.site.login = lambda *args, **kwargs: admin_login(
    admin.site, *args, **kwargs
)
# default: allow only non faked user with superuser and staff permissions
admin.site.has_permission = lambda *args, **kwargs: get_settings_func(
    "HAS_ADMIN_PERMISSION_FUNC",
    "spkcspider.apps.spider.functions.has_admin_permission"
)(admin.site, *args, **kwargs)

urlpatterns = [
    path('admin/', admin.site.urls),
    path(
        '',
        ComponentPublicIndex.as_view(
            is_home=True, template_name="spider_base/home.html"
        ),
        name="home"
    ),
    path('i18n/', include('django.conf.urls.i18n'))
]

for app in apps.get_app_configs():
    url_path = getattr(
        app, "spider_url_path",
        None
    )
    if not url_path:
        continue
    urlpatterns.append(
        path(
            url_path,
            include("{}.urls".format(app.name))
        )
    )


if getattr(settings, "SPIDER_LEGACY_REDIRECT", False):
    urlpatterns.insert(
        0,
        path(
            'spider/content/<int:token1>/<str:token2>/<slug:access>/',
            RedirectView.as_view(
                url='/spider/content/%(token1)s_%(token2)s/%(access)s/',
                permanent=True
            )
        )
    )
    urlpatterns.insert(
        0,
        path(
            'spider/components/<int:token1>/<str:token2>/list/',
            RedirectView.as_view(
                url='/spider/components/%(token1)s_%(token2)s/list/',
                permanent=True
            )
        )
    )
    urlpatterns.insert(
        0,
        path(
            'spider/content/access/<int:token1>/<str:token2>/<slug:access>/',
            RedirectView.as_view(
                url='/spider/content/%(token1)s_%(token2)s/%(access)s/',
                permanent=True
            )
        )
    )
    urlpatterns.insert(
        0,
        path(
            'spider/ucs/list/<int:token1>/<str:token2>/',
            RedirectView.as_view(
                url='/components/%(token1)s_%(token2)s/list/',
                permanent=True
            )
        )
    )


if getattr(settings, "USE_CAPTCHAS", False):
    urlpatterns.append(path(r'captcha/', include('captcha.urls')))

if 'django.contrib.flatpages' in settings.INSTALLED_APPS:
    urlpatterns.append(
        path('pages/', include('django.contrib.flatpages.urls'))
    )

urlpatterns += [
    # daily
    path('favicon.ico', cache_page(86400)(favicon_view)),
    path('robots.txt', cache_page(86400)(robots_view)),
    # daily
    path(
        'sitemap.xml',
        cache_page(86400)(
            sitemaps_views.index
        ),
        {
            'sitemaps': sitemaps,
            'sitemap_url_name': 'sitemaps'
        },
        name='django.contrib.sitemaps.views.index'
    ),
    # hourly
    path(
        'sitemap-<section>.xml',
        cache_page(3600)(
            sitemaps_views.sitemap
        ),
        {'sitemaps': sitemaps},
        name='sitemaps'
    ),
]


if settings.DEBUG:
    urlpatterns += static(
        settings.MEDIA_URL, document_root=settings.MEDIA_ROOT
    )



# Generated by Django 3.0.1 on 2019-12-30 13:00

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields
import spkcspider.apps.spider.models.content_extended
import spkcspider.constants.protections


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0009_auto_20190929_2117'),
    ]

    operations = [
        migrations.CreateModel(
            name='M2MTravelProtContent',
            fields=[
                ('id', models.BigAutoField(editable=False,
                                           primary_key=True, serialize=False)),
                ('source', models.ForeignKey(limit_choices_to=models.Q(('ctype__name', 'TravelProtection'), ('ctype__name', 'SelfProtection'),
                                                                       _connector='OR'), on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent')),
                ('target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                             related_name='+', to='spider_base.AssignedContent')),
            ],
        ),
        migrations.CreateModel(
            name='M2MTravelProtComponent',
            fields=[
                ('id', models.BigAutoField(editable=False,
                                           primary_key=True, serialize=False)),
                ('source', models.ForeignKey(limit_choices_to=models.Q(('ctype__name', 'TravelProtection'), ('ctype__name', 'SelfProtection'),
                                                                       _connector='OR'), on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent')),
                ('target', models.ForeignKey(limit_choices_to={
                 'strength__lt': 10}, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.UserComponent')),
            ],
        ),
        migrations.CreateModel(
            name='AttachedBlob',
            fields=[
                ('id', models.BigAutoField(editable=False,
                                           primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, default='', max_length=50)),
                ('unique', models.BooleanField(blank=True, default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('blob', models.BinaryField(blank=True, default=b'', editable=True)),
            ],
        ),
        migrations.CreateModel(
            name='AttachedCounter',
            fields=[
                ('id', models.BigAutoField(editable=False,
                                           primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, default='', max_length=50)),
                ('unique', models.BooleanField(blank=True, default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('counter', models.BigIntegerField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AttachedFile',
            fields=[
                ('id', models.BigAutoField(editable=False,
                                           primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, default='', max_length=50)),
                ('unique', models.BooleanField(blank=True, default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('file', models.FileField(upload_to=spkcspider.apps.spider.models.content_extended.get_file_path)),
            ],
        ),
        migrations.CreateModel(
            name='AttachedTimeSpan',
            fields=[
                ('id', models.BigAutoField(editable=False,
                                           primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, default='', max_length=50)),
                ('unique', models.BooleanField(blank=True, default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('start', models.DateTimeField()),
                ('stop', models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DataContent',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('quota_data', jsonfield.fields.JSONField(default=dict, blank=True)),
                ('free_data', jsonfield.fields.JSONField(default=dict, blank=True)),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
            managers=[
                ('objects', spkcspider.apps.spider.models.content_extended.DataContentManager()),
            ],
        ),
        migrations.AlterField(
            model_name='assignedprotection',
            name='state',
            field=models.CharField(choices=[(spkcspider.constants.protections.ProtectionStateType['disabled'], 'disabled'), (spkcspider.constants.protections.ProtectionStateType['enabled'], 'enabled'), (spkcspider.constants.protections.ProtectionStateType['instant_fail'], 'instant fail')], default=spkcspider.constants.protections.ProtectionStateType['enabled'], help_text='State of the protection.', max_length=1),
        ),
        migrations.AlterUniqueTogether(
            name='assignedcontent',
            unique_together=set(),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='protect_components',
            field=models.ManyToManyField(blank=True, related_name='travel_protected',
                                         through='spider_base.M2MTravelProtComponent', through_fields=("source", "target"), to='spider_base.UserComponent'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='protect_contents',
            field=models.ManyToManyField(blank=True, related_name='travel_protected',
                                         through='spider_base.M2MTravelProtContent', through_fields=("source", "target"), to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='datacontent',
            name='associated',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='attachedtimespan',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                    related_name='attachedtimespans', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='attachedfile',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                    related_name='attachedfiles', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='attachedcounter',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                    related_name='attachedcounters', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='attachedblob',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                    related_name='attachedblobs', to='spider_base.AssignedContent'),
        ),
        migrations.RemoveField(
            model_name='assignedcontent',
            name='content_type',
        ),
        migrations.AddConstraint(
            model_name='attachedblob',
            constraint=models.UniqueConstraint(condition=models.Q(
                unique=True), fields=('content', 'name'), name='attachedblob_unique'),
        ),
        migrations.AddConstraint(
            model_name='attachedcounter',
            constraint=models.UniqueConstraint(condition=models.Q(unique=True), fields=(
                'content', 'name'), name='attachedcounter_unique'),
        ),
        migrations.AddConstraint(
            model_name='attachedfile',
            constraint=models.UniqueConstraint(condition=models.Q(
                unique=True), fields=('content', 'name'), name='attachedfile_unique'),
        ),
        migrations.AddConstraint(
            model_name='attachedtimespan',
            constraint=models.UniqueConstraint(condition=models.Q(unique=True), fields=(
                'content', 'name'), name='attachedtimespan_unique'),
        ),
        migrations.AddConstraint(
            model_name='attachedtimespan',
            constraint=models.CheckConstraint(check=models.Q(('start__lte', django.db.models.expressions.F('stop')), ('start__isnull', True), ('stop__isnull', True), _connector='OR'), name='attachedtimespan_order'),
        ),
        migrations.AddConstraint(
            model_name='attachedtimespan',
            constraint=models.CheckConstraint(check=models.Q(('start__isnull', False), ('stop__isnull', False), _connector='OR'), name='attachedtimespan_exist'),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-04 17:38

from django.db import migrations


def fix_usertags(apps, schema_editor):
    DataContent = apps.get_model("spider_base", "DataContent")
    TagLayout = apps.get_model("spider_tags", "TagLayout")

    for d in DataContent.objects.filter(
        associated__ctype__name="TagLayout"
    ):
        layout = TagLayout.objects.get(id=d.free_data["tmp_layout_id"])
        layout.usertag = d.associated
        layout.save()
        del d.free_data["tmp_layout_id"]
        d.clean()
        d.save(update_fields=["free_data"])


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0010_auto_20200104_1737'),
    ]

    operations = [
        migrations.RunPython(fix_usertags),
    ]



# Generated by Django 3.0.1 on 2019-12-30 13:51

from django.db import migrations
import spkcspider.apps.spider_filets.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0013_auto_20191230_1351'),
    ]

    operations = [
        migrations.DeleteModel(
            name='FileFilet',
        ),
        migrations.DeleteModel(
            name='TextFilet',
        ),
        migrations.CreateModel(
            name='FileFilet',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=(spkcspider.apps.spider_filets.models.LicenseMixin, 'spider_base.datacontent'),
        ),
        migrations.CreateModel(
            name='TextFilet',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=(spkcspider.apps.spider_filets.models.LicenseMixin, 'spider_base.datacontent'),
        ),
    ]



__all__ = ["SpiderFiletsConfig"]

from django.apps import AppConfig


class SpiderFiletsConfig(AppConfig):
    name = 'spkcspider.apps.spider_filets'
    label = 'spider_filets'
    verbose_name = 'spkcspider File and Text content'



# Generated by Django 3.0.1 on 2019-12-30 13:05

from django.db import migrations


def move_to_datacontent(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    LinkContent = apps.get_model("spider_base", "LinkContent")
    TravelProtection = apps.get_model("spider_base", "TravelProtection")
    DataContent = apps.get_model("spider_base", "DataContent")

    for a in AssignedContent.objects.filter(
        ctype__name="Link"
    ):
        d = DataContent(associated=a)
        content = LinkContent.objects.get(id=a.object_id)
        d.free_data["push"] = content.push
        d.save()

    for assigned in AssignedContent.objects.filter(
        ctype__code=TravelProtection._meta.model_name
    ):
        DataContent.objects.create(associated=assigned)
        content = TravelProtection.objects.get(id=assigned.object_id)
        if content.active:
            assigned.info += "active\x1e"

        assigned.info += \
            "travel_protection_type={}\x1e".format(content.login_protection)
        if content.start or content.stop:
            assigned.attachedtimespans.create(
                unique=False,
                name="active",
                start=content.start,
                stop=content.stop
            )
        assigned.protect_contents.set(content.protect_contents.all())
        assigned.protect_components.set(content.protect_components.all())
        assigned.save()


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0010_auto_20191230_1300'),
    ]

    operations = [
        migrations.RunPython(move_to_datacontent),
    ]



"""
Base for Contents
namespace: spider_base

"""

__all__ = [
    "ContentVariant", "AssignedContent"
]

import logging
from base64 import b64encode

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from django.apps import apps
from django.conf import settings
from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from spkcspider.constants import (
    MAX_TOKEN_B64_SIZE, TravelProtectionType, VariantType,
    hex_size_of_bigid, static_token_matcher, travel_scrypt_params,
    ProtectionStateType
)
from spkcspider.utils.security import create_b64_id_token

from .. import registry
from ..abstract_models import BaseInfoModel, BaseSubUserModel
from ..queryfilters import travelprotection_types_q
from ..validators import content_name_validator, validator_token

logger = logging.getLogger(__name__)


login_choices_dict = dict(TravelProtectionType.as_choices())


# ContentType is already occupied
class ContentVariant(models.Model):
    id: int = models.BigAutoField(primary_key=True, editable=False)
    # usercontent abilities/requirements
    ctype: str = models.CharField(
        max_length=10, default="",
    )
    code: str = models.CharField(max_length=255)
    name: str = models.SlugField(max_length=50, unique=True, db_index=True)
    # required protection strength (selection)
    strength: int = models.PositiveSmallIntegerField(
        default=0, validators=[validators.MaxValueValidator(10)]
    )
    # for content features
    valid_feature_for = models.ManyToManyField(
        "self", blank=True, related_name="valid_features", symmetrical=False
    )

    objects = models.Manager()

    @property
    def installed_class(self):
        return registry.contents[self]

    @property
    def feature_urls(self):
        return registry.feature_urls[self]

    @property
    def is_feature(self):
        return (
            VariantType.component_feature.value in self.ctype or
            VariantType.content_feature.value in self.ctype
        )

    @property
    def deletion_period(self):
        return registry.content_deletion_periods[self]

    def localize_name(self):
        if self not in registry.contents:
            return self.name
        return registry.contents[self].localize_name(self.name)

    def __str__(self):
        return self.localize_name()

    def __repr__(self):
        return "<ContentVariant: %s>" % self.__str__()


class TravelProtectionManager(models.Manager):
    def get_active(self, now=None):
        if not now:
            now = timezone.now()
        AttachedTimespan = apps.get_model("spider_base", "AttachedTimespan")
        timespans = AttachedTimespan.objects.filter(
            models.Q(
                name="active",
                start__isnull=True,
                stop__gte=now
            ) |
            models.Q(
                name="active",
                start__lte=now,
                stop__isnull=True
            ) |
            models.Q(
                name="active",
                start__lte=now,
                stop__gte=now
            )
        )
        q = models.Q(info__contains="\x1eactive\x1e")
        q &= travelprotection_types_q
        q &= (
            models.Q(attachedtimespans__in=timespans) |
            models.Q(
                ctype__name="SelfProtection"
            )
        )
        return self.filter(q)

    def get_active_for_session(self, session, user, now=None):
        q = ~models.Q(info__contains="\x1epwhash=")
        if user.is_authenticated:
            for pw in session.get("travel_hashed_pws", []):
                q |= models.Q(
                    info__contains="\x1epwhash=%s\x1e" % pw
                )
            q &= models.Q(usercomponent__user=user)
        return self.get_active(now).filter(q)

    def get_active_for_request(self, request, now=None):
        return self.get_active_for_session(request.session, request.user)

    def handle_disable(self, travelprotection, request, now):
        return False

    def handle_trigger_hide(self, travelprotection, request, now):
        travelprotection.replace_info(
            travel_protection_type=TravelProtectionType.hide,
            pwhash=False
        )
        travelprotection.save(update_fields=["info"])
        return True

    def handle_trigger_disable(self, travelprotection, request, now):
        travelprotection.replace_info(
            travel_protection_type=TravelProtectionType.disable,
            pwhash=False
        )
        travelprotection.save(update_fields=["info"])
        return False

    def handle_wipe(self, travelprotection, request, now):
        # first components have to be deleted
        travelprotection.protect_components.all().delete()
        # as this deletes itself and therefor the information
        # about affected components
        travelprotection.protect_contents.all().delete()
        return True

    def handle_trigger_disable_user(self, travelprotection, request, now):
        AssignedProtection = \
            apps.get_model("spider_base", "AssignedProtection")
        UserComponent = \
            apps.get_model("spider_base", "UserComponent")
        travel = self.model.travel.get_active(now)
        q = models.Q(name="index")
        if travelprotection.content.free_data.get(
            "is_travel_protected", False
        ):
            q |= models.Q(travel_protected__in=travel)
        ucs = UserComponent.objects.select_for_update().exclude(q)
        ps = AssignedProtection.objects.select_for_update().filter(
            usercomponent__in=ucs
        )
        with transaction.atomic():
            ucs.update(
                strength=9, public=False
            )
            ucs.filter(required_passes=0).update(required_passes=1)
            ps.update(state=ProtectionStateType.disabled)
        travelprotection.replace_info(
            travel_protection_type=TravelProtectionType.disable,
            pwhash=False
        )
        travelprotection.save(update_fields=["info"])
        return False

    def handle_wipe_user(self, travelprotection, request, now):
        if travelprotection.content.free_data.get(
            "is_travel_protected", False
        ):
            return self.handle_trigger_disable_user(
                travelprotection, request, now
            )
        else:
            travelprotection.usercomponent.user.delete()
        return False

    _action_map = {
        val: "handle_%s" % key
        for key, val in TravelProtectionType.__members__.items()
    }

    def auth(self, request, uc, now=None):
        if not now:
            now = timezone.now()
        active = self.get_active(now).filter(
            usercomponent=uc
        )

        request.session["travel_hashed_pws"] = list(map(
            lambda x: b64encode(Scrypt(
                salt=settings.SECRET_KEY.encode("utf-8"),
                backend=default_backend(),
                **travel_scrypt_params
            ).derive(x[:128].encode("utf-8"))).decode("ascii"),
            request.POST.getlist("password")[:4]
        ))

        q = ~models.Q(info__contains="\x1epwhash=")
        for pw in request.session["travel_hashed_pws"]:
            q |= models.Q(
                info__contains="\x1epwhash=%s\x1e" % pw
            )

        active = active.filter(q)

        request.session["is_travel_protected"] = active.exists()

        # don't query further, use cached result (is_travel_protected)
        if not request.session["is_travel_protected"]:
            return True

        can_continue = True

        for travelprotection in active:
            travel_protection_type = \
                travelprotection.getlist("travel_protection_type", amount=1)[0]
            ret = getattr(
                self,
                self._action_map[travel_protection_type],
                lambda _1, _2, _3: True
            )(travelprotection, request, now)
            if not ret:
                can_continue = False
        return can_continue


class AssignedContentManager(models.Manager):
    """
        Manager for AssignedContent.
        Contains some base functionality for easing usage
        is not available in migrations
    """

    def create(self, **kwargs):
        ret = self.get_queryset().create(**kwargs)
        if not ret.token:
            ret.token = create_b64_id_token(ret.id, "_")
            ret.save(update_fields=["token"])
        return ret

    def update_or_create(self, **kwargs):
        ret = self.get_queryset().update_or_create(**kwargs)
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def get_or_create(self, defaults=None, **kwargs):
        ret = self.get_queryset().get_or_create(**kwargs)
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def from_token(
        self, token, *, info=frozenset(), variant=frozenset(),
        check_feature=False
    ):
        # check_feature requires that ctype of the selected content is in
        #  component features
        if isinstance(variant, str):
            variant = [variant]

        if isinstance(info, str):
            info = [info]

        q = models.Q()
        if variant:
            if isinstance(variant, models.QuerySet):
                q &= models.Q(ctype__in=variant)
            elif isinstance(variant, models.Model):
                q &= models.Q(ctype=variant)
            else:
                q &= models.Q(ctype__name__in=variant)
        for i in info:
            q &= models.Q(info__contains="\x1e%s\x1e" % i)
        if check_feature:
            q &= models.Q(
                ctype__feature_for_components__authtokens=token
            )
        return self.get(
            q, attached_to_token=token
        )

    def from_url_part(
        self, url, *, info=frozenset(), variant=frozenset(),
        check_feature=False
    ):
        """
            url: can be full url or token/accessmethod
            info: should be either string or iterable which will be matched
                  against info to retrieve an unique content
            returns: (<matched content/feature>, <content>/None)
        """
        UserComponent = apps.get_model("spider_base", "UserComponent")
        res = static_token_matcher.match(url)
        if not res:
            raise self.model.DoesNotExist()
        res = res.groupdict()
        # shortcut, we don't need to look up db to see that without
        # info field multiple items can match easily
        # the special case: only one content is unreliable and should
        # never tried to match
        if not info and not variant:
            raise self.model.MultipleObjectsReturned()

        if isinstance(variant, str):
            variant = [variant]

        if isinstance(info, str):
            info = [info]

        q = models.Q()
        if variant:
            if isinstance(variant, models.QuerySet):
                q &= models.Q(ctype__in=variant)
            elif isinstance(variant, models.Model):
                q &= models.Q(ctype=variant)
            else:
                q &= models.Q(ctype__name__in=variant)
        for i in info:
            q &= models.Q(info__contains="\x1e%s\x1e" % i)
        if res["access"] == "list":
            uc = UserComponent.objects.filter(
                token=res["static_token"]
            ).first()
            if check_feature:
                q &= (
                    models.Q(
                        ctype__feature_for_components=uc
                    )
                )
            return (self.get(q, usercomponent=uc), None)
        elif check_feature:
            content = self.get(
                token=res["static_token"]
            )
            q &= (
                models.Q(
                    ctype__feature_for_contents=content
                ) |
                models.Q(
                    ctype__feature_for_components__contents=content
                )
            )
            return (self.get(q), content)
        else:
            content = self.get(
                q, token=res["static_token"]
            )
            return (content, content)


class M2MTravelProtContent(models.Model):
    id: int = models.BigAutoField(primary_key=True, editable=False)
    source = models.ForeignKey(
        "spider_base.AssignedContent", related_name="+",
        on_delete=models.CASCADE, limit_choices_to=travelprotection_types_q
    )
    target = models.ForeignKey(
        "spider_base.AssignedContent", related_name="+",
        on_delete=models.CASCADE
    )


class M2MTravelProtComponent(models.Model):
    id: int = models.BigAutoField(primary_key=True, editable=False)
    source = models.ForeignKey(
        "spider_base.AssignedContent", related_name="+",
        on_delete=models.CASCADE, limit_choices_to=travelprotection_types_q
    )
    target = models.ForeignKey(
        "spider_base.UserComponent", related_name="+",
        on_delete=models.CASCADE, limit_choices_to={
            "strength__lt": 10,
        }
    )


class AssignedContent(BaseInfoModel, BaseSubUserModel):
    id: int = models.BigAutoField(primary_key=True, editable=False)
    # fake_level = models.PositiveIntegerField(null=False, default=0)
    attached_to_token = models.ForeignKey(
        "spider_base.AuthToken", blank=True, null=True,
        on_delete=models.CASCADE
    )
    # don't use extensive recursion,
    # this can cause performance problems and headaches
    # this is not enforced for allowing some small chains
    # see SPIDER_MAX_EMBED_DEPTH setting for limits (default around 5)
    attached_to_content = models.ForeignKey(
        "self", blank=True, null=True,
        related_name="attached_contents", on_delete=models.CASCADE
    )
    # set to indicate creating a new token
    token_generate_new_size = None
    # brute force protection and identifier, replaces nonce
    #  16 = usercomponent.id in hexadecimal
    #  +1 for seperator
    token: str = models.CharField(
        max_length=(MAX_TOKEN_B64_SIZE)+hex_size_of_bigid+2,
        db_index=True, unique=True, null=True, blank=True,
        validators=[validator_token]
    )
    # regex disables controlcars and disable special spaces
    # and allows some of special characters
    name: str = models.CharField(
        max_length=255, blank=True, default="",
        validators=[
            content_name_validator
        ]
    )
    description: str = models.TextField(
        default="", blank=True
    )
    usercomponent = models.ForeignKey(
        "spider_base.UserComponent", on_delete=models.CASCADE,
        related_name="contents", null=False, blank=False
    )
    features = models.ManyToManyField(
        "spider_base.ContentVariant",
        related_name="feature_for_contents", blank=True,
        limit_choices_to=models.Q(
            ctype__contains=VariantType.content_feature
        )
    )
    # ctype is here extended: VariantObject with abilities, name, model_name
    ctype = models.ForeignKey(
        ContentVariant, editable=False, null=True,
        on_delete=models.SET_NULL
    )
    priority: int = models.SmallIntegerField(default=0, blank=True)

    # creator = models.ForeignKey(
    #    settings.AUTH_USER_MODEL, editable=False, null=True,
    #    on_delete=models.SET_NULL
    # )
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)
    # only editable for admins
    deletion_requested = models.DateTimeField(
        null=True, blank=True, default=None
    )
    # required protection strength (real)
    strength = models.PositiveSmallIntegerField(
        default=0, validators=[validators.MaxValueValidator(10)],
        editable=False
    )
    # required protection strength for links, 11 to disable links
    strength_link = models.PositiveSmallIntegerField(
        default=0, validators=[validators.MaxValueValidator(11)],
        editable=False
    )
    attached_to_primary_anchor = models.BooleanField(
        default=False, editable=False, null=False,
        help_text=_(
            "Content references primary anchor"
        )
    )
    # - for a quick retrieval of related contents !!
    # - includes attached_contents
    # - is used by embed to find and output relations
    # - with get_forward_references can additional contents be added in embed
    references = models.ManyToManyField(
        "self", related_name="referenced_by", editable=False,
        symmetrical=False
    )

    protect_components = models.ManyToManyField(
        "spider_base.UserComponent",
        through=M2MTravelProtComponent, through_fields=("source", "target"),
        related_name="travel_protected",
        blank=True
    )
    protect_contents = models.ManyToManyField(
        "self", symmetrical=False,
        through=M2MTravelProtContent, through_fields=("source", "target"),
        related_name="travel_protected",
        blank=True
    )

    # info extra flags:
    #  primary: primary content of type for usercomponent
    #  unlisted: not listed
    #  anchor:
    objects = AssignedContentManager()
    travel = TravelProtectionManager()

    class Meta:
        constraints = []
        if (
            settings.DATABASES["default"]["ENGINE"] !=
            "django.db.backends.mysql"
        ):
            constraints.append(
                models.UniqueConstraint(
                    fields=['usercomponent', 'info'], name='unique_info'
                )
            )

    def __init__(self, *args, **kwargs):
        if isinstance(kwargs.get("ctype"), str):
            kwargs["ctype"] = ContentVariant.objects.get(name=kwargs["ctype"])
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<AssignedContent: ({}: {})>".format(
            self.usercomponent.username,
            self.name
        )

    def get_absolute_url(self, scope="view"):
        return reverse(
            "spider_base:ucontent-access",
            kwargs={"token": self.token, "access": scope}
        )

    # use next_object, last_object instead get_next_by_FOO, ...
    # for preventing disclosure of elements

    def get_size(self):
        return self.content.get_size()

    def localized_description(self):
        """ localize description either with own version or the localized """
        if not self.content:
            return self.description
        return self.content.localized_description()

    def clean(self):
        _ = gettext
        if VariantType.persist in self.ctype.ctype:
            if not self.attached_to_token:
                raise ValidationError(
                    _('Persistent token required'),
                    code="persist",
                )

        if not self.usercomponent.user_info.allowed_content.filter(
            name=self.ctype.name
        ).exists():  # pylint: disable=no-member
            raise ValidationError(
                message=_('Not an allowed ContentVariant for this user'),
                code='disallowed_contentvariant'
            )
        if self.strength > self.usercomponent.strength:
            raise ValidationError(
                _('Protection strength too low, required: %(strength)s'),
                code="strength",
                params={'strength': self.strength},
            )
        if (
            settings.DATABASES["default"]["ENGINE"] ==
            "django.db.backends.mysql"
        ):
            obj = AssignedContent.objects.filter(
                usercomponent=self.usercomponent,
                info=self.info
            ).first()

            if obj and obj.id != getattr(self, "id", None):
                raise ValidationError(
                    message=_("Unique Content already exists"),
                    code='unique_together',
                )
        super().clean()

    @cached_property
    def content(self):
        return self.ctype.installed_class.objects.get(associated=self)

    @property
    def is_hidden(self):
        return self.getflag("unlisted")

    @property
    def deletion_period(self):
        return self.ctype.deletion_period



# NOT IMPORTED BY DEFAULT
# especially because of gettext_lazy import

__all__ = [
    "VERIFICATION_CHOICES"
]

from django.utils.translation import gettext_lazy as _


VERIFICATION_CHOICES = [
    ("pending", _("pending")),
    ("verified", _("verified")),
    ("invalid", _("invalid")),
    ("rejected", _("rejected")),
]




__all__ = [
    "BaseSubUserModel"
]

from django.db import models


class BaseSubUserModel(models.Model):

    class Meta:
        abstract = True

    @property
    def user(self):
        return self.usercomponent.user  # pylint: disable=no-member

    @property
    def username(self):
        return self.usercomponent.username  # pylint: disable=no-member



# Generated by Django 3.0.2 on 2020-01-04 17:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0009_auto_20200104_1736'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taglayout',
            name='usertag',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                       related_name='+', to='spider_base.AssignedContent'),
        ),
        migrations.DeleteModel(
            name='UserTagLayout',
        ),
        migrations.CreateModel(
            name='UserTagLayout',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('spider_base.datacontent',),
        ),
    ]



# Generated by Django 3.0.1 on 2019-12-30 13:51

from django.db import migrations


def move_to_datacontent(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    TextFilet = apps.get_model("spider_filets", "TextFilet")
    FileFilet = apps.get_model("spider_filets", "FileFilet")
    DataContent = apps.get_model("spider_base", "DataContent")

    for a in AssignedContent.objects.filter(
        ctype__code=TextFilet._meta.model_name
    ):
        d = DataContent(associated=a)
        content = TextFilet.objects.get(id=a.object_id)
        d.free_data["push"] = content.push
        d.free_data["editable_from"] = list(content.editable_from.values_list(
            "id", flat=True
        ))
        d.free_data["license_name"] = content.license_name
        d.quota_data["license_url"] = content.license_url
        d.quota_data["sources"] = content.sources
        d.save()
        a.attachedblobs.create(
            unique=True, name="text", blob=content.text.encode("utf8")
        )
    for a in AssignedContent.objects.filter(
        ctype__code=FileFilet._meta.model_name
    ):
        d = DataContent(associated=a)
        content = FileFilet.objects.get(id=a.object_id)
        d.free_data["license_name"] = content.license_name
        d.quota_data["license_url"] = content.license_url
        d.quota_data["sources"] = content.sources
        d.save()
        a.attachedfiles.create(unique=True, name="file", file=content.file)


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0011_auto_20191230_1305'),
        ('spider_filets', '0012_auto_20190419_0800'),
    ]
    run_before = [
        ('spider_base', '0012_auto_20191230_1305'),
    ]

    operations = [
        migrations.RunPython(move_to_datacontent),
    ]



__all__ = ["SpiderBaseConfig"]

from django.apps import AppConfig
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.signals import (
    m2m_changed, post_delete, post_migrate, post_save, pre_save
)

from .signals import (
    CleanupCb, InitUserCb, TriggerUpdate, UpdateAnchorComponentCb,
    FeaturesCb, UpdateContentCb, UpdateSpiderCb, update_dynamic,
    DeleteFilesCb
)


class SpiderBaseConfig(AppConfig):
    name = 'spkcspider.apps.spider'
    label = 'spider_base'
    spider_url_path = 'spider/'
    spider_protections_path = ".protections"
    verbose_name = 'spkcspider base'

    def ready(self):
        from .models import (
            AssignedContent, UserComponent, AttachedFile
        )

        #######################
        post_save.connect(
            InitUserCb, sender=get_user_model(),
        )

        update_dynamic.connect(
            UpdateSpiderCb,
            dispatch_uid="update_spider_dynamic"
        )

        if getattr(settings, "UPDATE_DYNAMIC_AFTER_MIGRATION", True):
            post_migrate.connect(
                TriggerUpdate, sender=self,
            )

        #########################

        pre_save.connect(
            UpdateAnchorComponentCb, sender=UserComponent,
        )
        post_save.connect(
            UpdateContentCb, sender=AssignedContent,
        )
        post_save.connect(
            FeaturesCb, sender=UserComponent
        )
        post_save.connect(
            FeaturesCb, sender=AssignedContent
        )
        m2m_changed.connect(
            FeaturesCb, sender=UserComponent.features.through
        )
        m2m_changed.connect(
            FeaturesCb, sender=AssignedContent.features.through,
        )

        post_delete.connect(
            DeleteFilesCb, sender=AttachedFile
        )

        # order important for the next two events
        post_delete.connect(
            CleanupCb, sender=UserComponent,
        )

        post_delete.connect(
            CleanupCb, sender=AssignedContent,
        )



# Generated by Django 3.0.1 on 2019-12-30 13:55

from django.db import migrations
import spkcspider.apps.spider_keys.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0005_auto_20191230_1355'),
    ]

    operations = [
        migrations.DeleteModel(
            name='AnchorKey',
        ),
        migrations.DeleteModel(
            name='AnchorServer',
        ),
        migrations.DeleteModel(
            name='PublicKey',
        ),
        migrations.CreateModel(
            name='AnchorKey',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=(spkcspider.apps.spider_keys.models.AnchorMixing, 'spider_base.datacontent'),
        ),
        migrations.CreateModel(
            name='AnchorServer',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=(spkcspider.apps.spider_keys.models.AnchorMixing, 'spider_base.datacontent'),
        ),
        migrations.CreateModel(
            name='PublicKey',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('spider_base.datacontent',),
        ),
    ]



__all__ = (
    "literalize", "field_to_python", "add_property", "add_by_field",
    "prepare_description"
)

import re
from urllib.parse import urljoin
from functools import reduce

from django.conf import settings
from django.forms import BoundField, Field
from rdflib import RDF, XSD, BNode, Literal, URIRef
from spkcspider.constants import spkcgraph

# for not spamming sets
_empty_set = frozenset()


def field_to_python(value):
    if isinstance(value, BoundField):
        data = value.initial
        if value.form.is_bound:
            value = value.field.bound_data(value.data, data)
        else:
            value = data
    return value


def literalize(
    ob=None, datatype=None, use_uriref=None, domain_base=""
):
    if isinstance(ob, BoundField):
        if not datatype:
            datatype = getattr(ob.field, "spkc_datatype", None)
        if use_uriref is None:
            use_uriref = getattr(ob.field, "spkc_use_uriref", None)
        ob = field_to_python(ob)
    elif isinstance(datatype, BoundField):
        if use_uriref is None:
            use_uriref = getattr(datatype.field, "spkc_use_uriref", None)
        datatype = getattr(datatype.field, "spkc_datatype", None)
    elif isinstance(datatype, Field):
        if use_uriref is None:
            use_uriref = getattr(datatype, "spkc_use_uriref", None)
        datatype = getattr(datatype, "spkc_datatype", None)
    if ob is None:
        return RDF.nil
    if hasattr(ob, "get_absolute_url"):
        if not datatype:
            # add raw=embed, download and hash downloaded object
            datatype = spkcgraph["hashableURI"]
        if use_uriref is None:
            use_uriref = True
        ob = ob.get_absolute_url()
    elif isinstance(ob, str) and not datatype:
        datatype = XSD.string
    if use_uriref:
        return URIRef(urljoin(domain_base, ob))
    return Literal(ob, datatype=datatype)


def add_property(
    graph, name, ref=None, ob=None, literal=None, datatype=None,
    iterate=False
):
    value_node = BNode()
    if ref:
        graph.add((
            ref, spkcgraph["properties"],
            value_node
        ))
    graph.add((
        value_node, spkcgraph["name"],
        Literal(name, datatype=XSD.string)
    ))
    if literal is None:
        literal = getattr(ob, name)
    if not iterate:
        literal = [literal]
    for l in literal:
        graph.add((
            value_node, spkcgraph["value"],
            Literal(l, datatype=datatype)
        ))
    if not literal:
        graph.set((value_node, spkcgraph["value"], RDF.nil))
    return value_node


def _field_helper(ob, attr):
    return getattr(ob, attr)


def add_by_field(dic, field="__name__"):
    def _func(klass):
        # should be able to block real object
        if "{}.{}".format(klass.__module__, klass.__qualname__) not in getattr(
            settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
        ):
            dic[reduce(_field_helper, [klass, *field.split(".")])] = klass

        return klass
    return _func


_cleanstr = re.compile(r'<+.*>+')
_whitespsplit = re.compile(r'\s+')


def prepare_description(raw_html, amount=0):
    text = _cleanstr.sub(' ', raw_html).\
        replace("<", " ").replace(">", " ").strip()
    return _whitespsplit.split(text, amount)



__all__ = [
    "TagLayoutForm", "TagLayoutAdminForm", "SpiderTagForm",
]

from django import forms
# from django.utils.translation import gettext_lazy as _
from django.conf import settings
# from django.apps import apps
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from spkcspider.apps.spider.fields import JsonField, MultipleOpenChoiceField
from spkcspider.apps.spider.models import (
    ReferrerObject
)
from spkcspider.apps.spider.widgets import ListWidget
from spkcspider.utils.urls import merge_get_url

from . import registry
from .models import SpiderTag, TagLayout
from .widgets import SchemeWidget

_extra = '' if settings.DEBUG else '.min'


class TagLayoutForm(forms.ModelForm):
    layout = JsonField(
        widget=SchemeWidget(registry.fields.keys())
    )
    default_verifiers = MultipleOpenChoiceField(
        widget=ListWidget(
            items={
                "format_type": "url"
            }, item_label=_("Url to Verifier")
        ), required=False
    )

    class Meta:
        model = TagLayout
        fields = ["name", "unique", "layout", "default_verifiers"]

    usertaglayout = None

    def __init__(self, usertaglayout, **kwargs):
        self.usertaglayout = usertaglayout
        super().__init__(**kwargs)

    def _save_m2m(self):
        self.instance.usertag = self.usertaglayout.associated
        self.instance.name = self.usertaglayout.associated.name
        self.instance.description = self.usertaglayout.associated.description
        self.instance.save()
        self.instance.usertag.refresh_from_db()
        return super()._save_m2m()

    def clean_default_verifiers(self):
        values = self.cleaned_data["default_verifiers"]
        if not isinstance(values, list):
            raise forms.ValidationError(
                _("Invalid format"),
                code='invalid_format'
            )
        values = list(map(merge_get_url, values))
        return values

    def clean(self):
        self.usertaglayout.full_clean()
        return super().clean()

    def save(self, commit=True):
        if commit:
            self.usertaglayout.save()
            self._save_m2m()
        else:
            self.save_m2m = self._save_m2m
        return self.usertaglayout


class TagLayoutAdminForm(forms.ModelForm):
    layout = JsonField(
        widget=SchemeWidget(registry.fields.keys())
    )
    default_verifiers = MultipleOpenChoiceField(
        widget=ListWidget(
            items={
                "format_type": "url"
            }, item_label=_("Url to Verifier")
        ), required=False
    )

    class Meta:
        model = TagLayout
        fields = ["name", "unique", "layout", "default_verifiers", "usertag"]

    class Media:
        css = {
            'all': [
                'node_modules/@fortawesome/fontawesome-free/css/all.min.css'
            ]
        }

    def clean_default_verifiers(self):
        values = self.cleaned_data["default_verifiers"]
        if not isinstance(values, list):
            raise forms.ValidationError(
                _("Invalid format"),
                code='invalid_format'
            )
        values = list(map(merge_get_url, values))
        return values


class SpiderTagForm(forms.ModelForm):
    updateable_by = MultipleOpenChoiceField(
        required=False, initial=False,
        widget=ListWidget(
            items={
                "format_type": "url"
            }, item_label=_("Url")
        )
    )
    layout = forms.ModelChoiceField(
        queryset=TagLayout.objects.none(),
        to_field_name="name"
    )

    class Meta:
        model = SpiderTag
        fields = ["layout", "updateable_by"]

    def __init__(self, user=None, **kwargs):
        super().__init__(**kwargs)
        index = user.usercomponent_set.get(name="index")
        self.fields["layout"].queryset = TagLayout.objects.filter(
            Q(usertag__isnull=True) |
            Q(usertag__usercomponent=index)
        ).order_by("name")

    def clean_updateable_by(self):
        values = self.cleaned_data["updateable_by"]
        values = set(map(merge_get_url, values))
        existing = ReferrerObject.objects.filter(
            url__in=values
        ).values_list("url", flat=True)
        for url in values.difference(existing):
            # extra validation
            ReferrerObject.objects.get_or_create(
                url=url
            )
        return ReferrerObject.objects.filter(
            url__in=values
        )



"""
namespace: spider_base

"""

#  __all__ = ("BaseProtection", "ProtectionResult")

import binascii
import functools
import ipaddress
import logging
from base64 import b64decode, b64encode
from hashlib import sha256
from random import SystemRandom

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

import ratelimit
from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth import authenticate
from django.http import Http404
from django.utils.crypto import constant_time_compare
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext
from django.views.decorators.debug import sensitive_variables
from ratelimit import parse_rate
from spkcspider.constants import ProtectionStateType, ProtectionType
from spkcspider.utils.fields import add_by_field
from spkcspider.utils.security import aesgcm_pbkdf2_cryptor, create_b64_token

from .fields import MultipleOpenChoiceField
from .widgets import ListWidget, OpenChoiceWidget, PWOpenChoiceWidget
from . import registry

# from django.contrib.auth.hashers import make_password


logger = logging.getLogger(__name__)
_sysrand = SystemRandom()

# don't spam set objects
_empty_set = frozenset()

# for debug/min switch
_extra = '' if settings.DEBUG else '.min'

_pbkdf2_params = {
    "iterations": 120000,
    "hash_name": "sha512",
    "dklen": 32
}

_Scrypt_params = {
    "length": 64,
    "n": 2**14,
    "r": 16,
    "p": 2,
}


def exclude_disabled_state(inp):
    # allow enabled and instant_fail
    return inp[0] != ProtectionStateType.disabled


@functools.lru_cache(maxsize=1)
def amount_pws():
    Protection = django_apps.get_model("spider_base", "Protection")
    return max(Protection.objects.filter(
        ptype__contains=ProtectionType.password
    ).count(), 2)


class ProtectionList(list):
    uses_password = False
    media = forms.Media()


class BaseProtection(forms.Form):
    """
        Base for Protections
        Usage: use form to define some configurable fields for configuration
        use auth to validate: in this case:
            template_name, render, form variable are used
    """
    use_required_attribute = False
    state = forms.ChoiceField(
        choices=ProtectionStateType.as_choices(),
        initial=ProtectionStateType.disabled,
        label=_("State"), required=False,
        widget=forms.RadioSelect(
            attrs={
                "class": "update-required_passes"
            }
        )
    )

    # unique code name max 10 slug chars
    # if imported by extract_app_dicts, name is automatically set to key name
    # name = foo

    # ptype valid for, is overwritten with current ptype
    ptype = ""

    # description of Protection
    description = None

    template_name = None
    # form for authentication
    form = None
    # optional render function
    # render = None

    # auto populated, instance
    protection = None
    instance = None
    parent_form = None

    # initial values
    initial = {}

    def __init__(self, *, protection, ptype, request, uc, form, **kwargs):
        AssignedProtection = django_apps.get_model(
            "spider_base", "AssignedProtection"
        )
        self.ptype = ptype
        self.parent_form = form
        self.instance = uc.protections.filter(
            protection=protection
        ).first()
        # here .value is required
        if not self.instance:
            # usercomponent can be partly initialized,
            #   don't rely on name or id in form __init__
            #   cleaning will be done after all usercomponent data is updated
            #   so there it is safe
            self.instance = AssignedProtection(
                usercomponent=uc, protection=protection,
                state=ProtectionStateType.disabled
            )
        initial = self.get_initial()
        if not self.instance.state:
            self.instance.state = ProtectionStateType.disabled
        # use instance informations
        initial["state"] = self.instance.state
        # copy of initial is saved as self.initial, so safe to change it
        # after __init__ is called
        super().__init__(initial=initial, **kwargs)
        self.fields["state"].help_text = self.description

    def get_initial(self):
        initial = self.initial.copy()
        initial.update(self.instance.data)
        return initial

    def get_strength(self):
        # can provide strength in range 0-4
        # 0 no protection
        # 1 weak protection
        # 2 normal protection
        # 3 strong protection
        # 4 reserved for login only, can be returned to auth user
        # tuple for min max
        return (1, 1)

    @staticmethod
    def extract_form_kwargs(request):
        kwargs = {}
        if request.method in ["POST", "PUT"]:
            kwargs = {
                'data': request.POST,
                'files': request.FILES,
            }
        return kwargs

    @classmethod
    def auth(cls, request, **kwargs):
        if hasattr(cls, "auth_form"):
            form = cls.auth_form(**cls.extract_form_kwargs(request))
            if form.is_valid():
                return 1
            return form
        return False

    @classmethod
    def localize_name(cls, name=None):
        if not name:
            name = cls.name
        return pgettext("protection name", name.title())

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name(name)

    @classmethod
    def get_auth_media(cls, result):
        if isinstance(result, forms.Form):
            return result.media
        elif hasattr(cls, "auth_form"):
            if result is None or isinstance(result, int):
                return cls.auth_form().media
            else:
                return cls.auth_form(**result).media
        else:
            return forms.Media()

    def __str__(self):
        return self.localize_name(self.name)

    def clean_state(self):
        s = self.cleaned_data.get("state", "")
        if not s:
            return ProtectionStateType.disabled
        return s

    def save(self):
        self.instance.state = self.cleaned_data.pop("state")
        self.instance.data = self.cleaned_data
        self.instance.save()


class PseudoPw(forms.Form):
    """ Pseudoclass, emulates Protection and AssignedProtection """
    # name clashes with the one of PasswordProtection
    # works because it is not used in db
    name = "password"
    code = "password"
    ptype = ""
    use_required_attribute = False

    password = MultipleOpenChoiceField(
        widget=ListWidget(
            items={
                "format_type": "password"
            },
            item_label=_("Password")
        ), required=False, initial=[""]
    )

    @property
    def installed_class(self):
        return self

    @classmethod
    def auth(cls, request, **kwargs):
        return cls()

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Password")

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name("Password")

    def __str__(self):
        return self.localize_name(self.name)


# only friends have access
@add_by_field(registry.protections, "name")
class FriendProtection(BaseProtection):
    name = "friends"
    ptype = ProtectionType.access_control

    users = MultipleOpenChoiceField(
        label=_("Users"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )
    description = _("Limit access to selected users")

    def get_strength(self):
        return (3, 3)

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if (
            obj and request.user.is_authenticated and
            (
                getattr(request.user, request.user.USERNAME_FIELD) in
                obj.data["users"]
            )
        ):
            return 3
        else:
            return False


@add_by_field(registry.protections, "name")
class RandomFailProtection(BaseProtection):
    name = "randomfail"
    ptype = (
        ProtectionType.access_control + ProtectionType.authentication +
        ProtectionType.side_effects
    )

    success_rate = forms.IntegerField(
        label=_("Success Rate"), min_value=20, max_value=100, initial=100,
        widget=forms.NumberInput(attrs={'type': 'range'}),
        help_text=_("Set success rate")
    )

    use_404 = forms.BooleanField(label="Use 404 errors?", required=False)

    description = _(
        "Fail/Refuse randomly. Optionally with 404 error, "
        "to disguise correct access."
    )

    def get_strength(self):
        return (0, 0)

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Random Fail")

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if obj and obj.data.get("success_rate", None):
            if _sysrand.randrange(1, 101) <= obj.data["success_rate"]:
                return 0
            elif obj.data.get("use_404", False):
                raise Http404()
        return False


@add_by_field(registry.protections, "name")
class RateLimitProtection(BaseProtection):
    name = "ratelimit"
    ptype = (
        ProtectionType.access_control + ProtectionType.authentication
    )

    rate_accessed = forms.RegexField(
        regex=r'([\d]+)/([\d]*)([smhd])?',
        required=False,
        help_text=_(
            "Maximal access tries to this component per ip "
            "before blocking. Format: tries/multiplier"
            "(smhdw, second till week)"
        )
    )
    rate_static_token_error = forms.IntegerField(
        required=False, min_value=1,
        help_text=_(
            "Maximal http404 errors per hour per user/ip before blocking"
        )
    )
    rate_login_failed_ip = forms.IntegerField(
        required=False, min_value=1,
        help_text=_(
            "Maximal failed logins per ip per hour before blocking"
        )
    )
    rate_login_failed_account = forms.IntegerField(
        required=False, min_value=1,
        help_text=_(
            "Maximal failed logins per hour for this user before blocking. "
            "Useful for login"
        )
    )

    description = _(
        "Ratelimit access tries"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if ProtectionType.authentication in self.ptype:
            pass

    def get_strength(self):
        return (1, 1)

    def clean_rate_accessed(self):
        try:
            ret = parse_rate(self.cleaned_data.get("rate_accessed"))
            if ret[0] <= 0 or ret[1] <= 0:
                return None
        except ValueError:
            return None
        return self.cleaned_data.get("rate_accessed")

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Rate Limit")

    @classmethod
    def count_access(cls, request, obj):
        return b"%b:%b" % (
            str(obj.usercomponent_id).encode("ascii"), ipaddress.ip_network(
                request.META['REMOTE_ADDR'], strict=False
            ).compressed.encode("ascii")
        )

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if not obj:
            return False
        if obj:
            temp = obj.data.get("rate_accessed", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_ratelimit_accessed",
                key=cls.count_access(request, obj), rate=temp, inc=True
            )["request_limit"] > 0:
                return False
            temp = obj.data.get("rate_static_token_error", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_static_token_error",
                key="user_or_ip", rate=(int(temp), 3600), inc=False
            )["request_limit"] > 0:
                return False
            temp = obj.data.get("rate_login_failed_ip", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_login_failed_ip",
                key="ip", rate=(int(temp), 3600), inc=False
            )["request_limit"] > 0:
                return False
            temp = obj.data.get("rate_login_failed_account", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_login_failed_account",
                key=lambda x, y: obj.usercomponent.username,
                rate=(int(temp), 3600), inc=False
            )["request_limit"] > 0:
                return False
        return 1


@add_by_field(registry.protections, "name")
class LoginProtection(BaseProtection):
    name = "login"
    ptype = (
        ProtectionType.authentication + ProtectionType.access_control +
        ProtectionType.password
    )

    description = _("Use Login password")

    allow_auth = forms.BooleanField(
        label=_("Component authentication"), required=False
    )
    _request = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._request = kwargs.get("request")
        if ProtectionType.authentication in self.ptype:
            self.fields["state"].initial = ProtectionStateType.enabled
            self.fields["state"].choices = \
                filter(
                    exclude_disabled_state, self.fields["state"].choices
                )
            del self.fields["allow_auth"]

    def clean(self):
        super().clean()
        if ProtectionType.authentication in self.ptype:
            if not authenticate(
                self._request, username=self.instance.usercomponent.username,
                password=self.parent_form.initial["master_pw"], nospider=True
            ):
                self.parent_form.add_error(
                    "master_pw",
                    forms.ValidationError(
                        _("Invalid Password"),
                        code="invalid_password"
                    )
                )
        return self.cleaned_data

    def get_strength(self):
        return (3, 4 if self.cleaned_data.get("allow_auth", False) else 3)

    @classmethod
    @sensitive_variables("password")
    def auth(cls, request, obj, **kwargs):
        if not obj:
            return False

        username = obj.usercomponent.username
        for password in request.POST.getlist("password")[:amount_pws()]:
            if authenticate(
                request, username=username, password=password, nospider=True
            ):
                if obj.data.get("allow_auth", False):
                    return 4
                return 3
        return False

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name("Password")


@add_by_field(registry.protections, "name")
class PasswordProtection(BaseProtection):
    # has same name as pseudo pw field,
    # works because it uses the password form and renders no own auth form
    name = "password"
    ptype = (
        ProtectionType.access_control + ProtectionType.authentication +
        ProtectionType.password
    )

    description = _("Protect with extra passwords")
    prefix = "protection_passwords"

    class Media:
        css = {
            'all': [
                'node_modules/@devkral/selectize/dist/css/selectize.default.css'  # noqa:E501
            ]
        }
        js = [
            'node_modules/jquery/dist/jquery%s.js' % _extra,
            'node_modules/@devkral/selectize/dist/js/standalone/selectize%s.js' % _extra,  # noqa: E501
            'node_modules/base64-js/base64js.min.js',
            'spider_base/protections/PasswordProtection.js'
        ]

    salt = forms.CharField(
        widget=forms.HiddenInput
    )
    default_master_pw = forms.CharField(
        widget=forms.HiddenInput, disabled=True
    )

    passwords = MultipleOpenChoiceField(
        label=_("Passwords"), required=False,
        widget=PWOpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    auth_passwords = MultipleOpenChoiceField(
        label=_("Passwords (for component authentication)"), required=False,
        widget=PWOpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    _successful_clean = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        salt = self.initial.get("salt")
        if not salt:
            salt = self.fields["salt"].widget.value_from_datadict(
                self.data, self.files, self.add_prefix("salt")
            )
        if not salt:
            salt = create_b64_token()
        self.initial["salt"] = salt
        self.initial["default_master_pw"] = sha256(
            "".join([salt, settings.SECRET_KEY]).encode("utf-8")
        ).hexdigest()
        if ProtectionType.authentication in self.ptype:
            del self.fields["auth_passwords"]

    @staticmethod
    def eval_strength(length):
        if not length:
            return 0
        elif length > 15:
            return 2
        elif length > 40:
            return 3
        return 1

    def get_strength(self):
        maxstrength = self.eval_strength(self.cleaned_data["max_length"])
        if len(self.cleaned_data["auth_passwords"]) > 0:
            maxstrength = 4
        return (
            self.eval_strength(self.cleaned_data["min_length"]),
            maxstrength
        )

    @classmethod
    def hash_pw(cls, pw, salt, params=_Scrypt_params):
        return b64encode(Scrypt(
            salt=salt,
            backend=default_backend(),
            **params
        ).derive(pw[:128].encode("utf-8"))).decode("ascii")

    def is_valid(self):
        if not self._successful_clean:
            return False
        return super().is_valid()

    def clean(self):
        super().clean()
        # prevents user self lockout
        if ProtectionType.authentication in self.ptype and \
           len(self.cleaned_data["passwords"]) == 0:
            self.cleaned_data["active"] = False
        elif (
            len(self.cleaned_data["passwords"]) == 0 and
            len(self.cleaned_data["auth_passwords"]) == 0
        ):
            self.cleaned_data["active"] = False

        if getattr(self.instance, "id", None) and not self.has_changed():
            self.cleaned_data.update(self.instance.data)
            self._successful_clean = True
            return self.cleaned_data

        # eliminate duplicates
        self.cleaned_data["passwords"] = list(set(
            self.cleaned_data.get("passwords", _empty_set)
        ))
        self.cleaned_data["auth_passwords"] = list(set(
            self.cleaned_data.get("auth_passwords", _empty_set)
        ))
        self.cleaned_data["pbkdf2_params"] = _pbkdf2_params

        min_length = None
        max_length = None
        cryptor = None
        salt = self.cleaned_data.get("salt", "").encode("ascii")
        if not salt:
            return self.cleaned_data
        master_pw = self.parent_form.initial.get("master_pw", None)
        self.cleaned_data.pop("default_master_pw", None)
        has_master_pw = True
        if not master_pw:
            has_master_pw = False
            master_pw = self.initial["default_master_pw"]
        cryptor = aesgcm_pbkdf2_cryptor(
            master_pw, salt=salt,
            params=self.cleaned_data["pbkdf2_params"]
        )

        hashed_passwords = []
        auth_passwords = []
        hashed_auth_passwords = []
        current_field = "passwords"
        try:
            for pw in self.cleaned_data["passwords"]:
                if pw.startswith("bogo"):
                    raise ValueError("bogo not allowed here")
                nonce, pw = map(b64decode, pw.split(":", 1))
                pw = cryptor.decrypt(nonce, pw, None).decode("utf-8")
                lenpw = len(pw)
                if not min_length or lenpw < min_length:
                    min_length = lenpw
                if not max_length or lenpw > max_length:
                    max_length = lenpw
                hashed_passwords.append(self.hash_pw(
                    pw, salt, params=_Scrypt_params
                ))

            current_field = "auth_passwords"

            for pw in self.cleaned_data["auth_passwords"]:
                if pw.startswith("bogo"):
                    raise ValueError("bogo not allowed here")
                pwsource = pw
                nonce, pw = map(b64decode, pw.split(":", 1))
                pw = cryptor.decrypt(nonce, pw, None).decode("utf-8")
                lenpw = len(pw)
                if self.eval_strength(lenpw) < 2:
                    continue
                if not min_length or lenpw < min_length:
                    min_length = lenpw
                if not max_length or lenpw > max_length:
                    max_length = lenpw
                auth_passwords.append(pwsource)
                hashed_auth_passwords.append(self.hash_pw(
                    pw, salt, params=_Scrypt_params
                ))

            self.cleaned_data["hashed_passwords"] = hashed_passwords
            self.cleaned_data["auth_passwords"] = auth_passwords
            self.cleaned_data["scrypt_params"] = _Scrypt_params
            self.cleaned_data["hashed_auth_passwords"] = hashed_auth_passwords
            self.cleaned_data["min_length"] = min_length
            self.cleaned_data["max_length"] = max_length
            self._successful_clean = True
        except InvalidTag:
            if has_master_pw:
                self.parent_form.add_error(
                    "master_pw",
                    forms.ValidationError(
                        _(
                            "Invalid master password? "
                            "Passwords were not readable"
                        )
                    )
                )
            else:
                self.add_error(
                    current_field,
                    forms.ValidationError(
                        _(
                            "Forgot to enter master password? "
                            "Passwords were not readable"
                        )
                    )
                )
        except (ValueError, binascii.Error) as exc:
            logger.warning("PWProtection: pw encoding problems", exc_info=exc)
            self.add_error(
                current_field,
                forms.ValidationError(
                    _("Format error, check passwords")
                )
            )

        return self.cleaned_data

    @classmethod
    @sensitive_variables("password", "pw")
    def auth(cls, request, obj, **kwargs):
        if not obj:
            return False
        success = False
        auth = False
        max_length = 0
        salt = obj.data.get("salt", "").encode("ascii")
        for password in request.POST.getlist("password")[:amount_pws()]:
            if salt:
                pwhash = cls.hash_pw(
                    password, salt, params=obj.data.get(
                        "scrypt_params", _Scrypt_params
                    )
                )

                for pw in obj.data["hashed_passwords"]:
                    if constant_time_compare(pw, pwhash):
                        success = True

                for pw in obj.data["hashed_auth_passwords"]:
                    if constant_time_compare(pw, pwhash):
                        success = True
                        auth = True
            else:

                for pw in obj.data["passwords"]:
                    if constant_time_compare(pw, password):
                        success = True

                for pw in obj.data.get("auth_passwords", []):
                    if constant_time_compare(pw, password):
                        success = True
                        auth = True
            if success:
                max_length = max(len(password), max_length)

        if success:
            if auth:
                return 4
            return cls.eval_strength(max_length)
        return False

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Extra Passwords")


if getattr(settings, "USE_CAPTCHAS", False):
    from captcha.fields import CaptchaField

    @add_by_field(registry.protections, "name")
    class CaptchaProtection(BaseProtection):
        name = "captcha"
        ptype = ProtectionType.access_control
        ptype += ProtectionType.authentication
        description = _("Require captcha")

        class auth_form(forms.Form):
            use_required_attribute = False
            prefix = "protection_captcha"

        auth_form.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
            CaptchaField(label=_("Captcha"))
        auth_form.base_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
            auth_form.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME]

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            if ProtectionType.authentication in self.ptype:
                if getattr(settings, "REQUIRE_LOGIN_CAPTCHA", False):
                    self.initial["state"] = \
                        ProtectionStateType.instant_fail
                    self.fields["state"].initial = \
                        ProtectionStateType.instant_fail
                    self.fields["state"].disabled = True
                    self.fields["instant_fail"].help_text = \
                        _("captcha is for login required (admin setting)")

        def get_strength(self):
            return (1, 1)

        @classmethod
        def localize_name(cls, name=None):
            return pgettext("protection name", "Captcha Protection")

        @classmethod
        def auth(cls, request, obj, **kwargs):
            if not obj:
                return cls.auth_form()
            form = cls.auth_form(**cls.extract_form_kwargs(request))
            if request.method != "GET" and form.is_valid():
                return 1
            return form



__all__ = ["SpiderTagsConfig"]

from django.apps import AppConfig
from spkcspider.apps.spider.signals import update_dynamic

from .signals import UpdateLayouts


class SpiderTagsConfig(AppConfig):
    name = 'spkcspider.apps.spider_tags'
    label = 'spider_tags'
    spider_url_path = 'spidertags/'
    spider_layouts_path = ".layouts"
    spider_fields_path = ".fields"
    verbose_name = 'spkcspider tags'

    def ready(self):
        update_dynamic.connect(
            UpdateLayouts,
            dispatch_uid="update_layouts"
        )



"""
Basic Contents like TravelProtection, or Links
namespace: spider_base

"""

__all__ = [
    "LinkContent", "PersistenceFeature", "DomainMode", "DefaultActions",
    "TravelProtection"
]

import logging

from django.shortcuts import redirect
from django.urls import reverse
from django.utils.html import escape
from django.utils.translation import gettext, pgettext

from spkcspider.constants import ActionUrl, VariantType
from spkcspider.utils.fields import add_by_field

from .. import registry
from ..abstract_models import BaseContent
from .content_extended import DataContent

logger = logging.getLogger(__name__)

# raw and export: use references
link_abilities = frozenset(["add", "update", "update_raw", "raw", "export"])


@add_by_field(registry.contents, "_meta.model_name")
class PersistenceFeature(BaseContent):
    appearances = [
        {
            "name": "Persistence",
            "ctype": VariantType.component_feature,
            "strength": 0
        },
    ]

    class Meta:
        abstract = True

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl(
                "renew-token",
                reverse("spider_base:token-renew")
            ),
        ]


@add_by_field(registry.contents, "_meta.model_name")
class DomainMode(BaseContent):
    appearances = [
        {
            "name": "DomainMode",
            "ctype": (
                VariantType.component_feature + VariantType.content_feature +
                VariantType.no_export
            ),
            "valid_feature_for": "*",
            "strength": 0
        },
    ]

    class Meta:
        abstract = True

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Domain Mode")


@add_by_field(registry.contents, "_meta.model_name")
class DefaultActions(BaseContent):
    appearances = [
        {
            "name": "DefaultActions",
            "ctype": (
                VariantType.component_feature + VariantType.no_export +
                VariantType.unlisted
            ),
            "strength": 0
        },
    ]

    @classmethod
    def feature_urls(cls, name):
        return registry.feature_urls.default_actions.items()

    class Meta:
        abstract = True


@add_by_field(registry.contents, "_meta.model_name")
class LinkContent(DataContent):
    appearances = [{
        "name": "Link",
        "ctype": VariantType.raw_update
    }]
    expose_name = False
    expose_description = False

    class Meta:
        proxy = True

    def get_abilities(self, context):
        ret = set()
        if self.associated.getflag("anchor"):
            ret.add("anchor")
        return ret

    def get_content_name(self):
        return self.associated.attached_to_content.content.get_content_name()

    def get_content_description(self):
        return (
            self.associated.
            attached_to_content.
            content.
            get_content_description()
        )

    def get_strength(self) -> int:
        return self.associated.attached_to_content.content.get_strength()

    def get_priority(self) -> int:
        priority = self.associated.attached_to_content.content.get_priority()
        # pin to top
        if self.free_data.get("push") and priority < 1:
            return 1
        return priority

    def get_strength_link(self) -> int:
        # don't allow links linking on links
        return 11

    def get_info(self):
        ret = self.associated.attached_to_content.content.get_info()
        return "%ssource=%s\x1elink\x1e" % (
            ret, self.associated_id
        )

    def get_form(self, scope):
        from ..forms import LinkForm
        if scope in link_abilities:
            return LinkForm
        # maybe not required
        return self.associated.attached_to_content.content.get_form(scope)

    def get_form_kwargs(self, **kwargs):
        if kwargs["scope"] in link_abilities:
            ret = super().get_form_kwargs(**kwargs)
            ret["uc"] = kwargs["uc"]
            ret["request"] = kwargs["request"]
        else:
            # maybe not required anymore
            ret = self.associated.attached_to_content.content.get_form_kwargs(
                **kwargs
            )
        return ret

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Create Content Link"))
        return super().access_add(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update Content Link"))
        return super().access_update(**kwargs)

    def access_raw_update(self, **kwargs):
        return redirect(
            'spider_base:ucontent-access',
            token=self.associated.attached_to_content.token,
            access='update'
        )

    def access(self, context):
        # context is updated and used outside!!
        if context["scope"] in link_abilities:
            return super().access(context)
        else:
            context["source"] = self
            context["uc"] = self.associated.attached_to_content.usercomponent
            ret = self.access(context)
            ret["uc"] = self.associated.usercomponent
            return ret


@add_by_field(registry.contents, "_meta.model_name")
class TravelProtection(DataContent):
    # should not be a feature as it would be detectable this way
    appearances = [
        {
            "name": "TravelProtection",
            "strength": 10,
        },
        {
            "name": "SelfProtection",
            "strength": 10,
        }
    ]

    force_token_size = 60
    expose_name = False
    expose_description = False

    _anonymous_deactivation = False
    _prepared_info = None

    class Meta:
        proxy = True
        permissions = [
            (
                "use_dangerous_travelprotections",
                "Can use dangerous TravelProtections"
            )
        ]

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        if name == "TravelProtection":
            return _("content name", "Travel-Protection")
        else:
            return _("content name", "Self-Protection")

    def get_strength_link(self):
        return 11

    def get_priority(self):
        # pin to top with higher priority
        return 2

    def get_form(self, scope):
        from ..forms import TravelProtectionForm
        return TravelProtectionForm

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["request"] = kwargs["request"]
        return ret

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Create Travel Protection"))
        # redirect, so user does not see token, needs msg foo for sending token
        # token = self.associated.token
        ret = super().access_add(**kwargs)
        # if (
        #     token != self.associated.token and
        #     TravelProtection.objects.get_active().filter(pk=self.pk).exists()
        # ):
        #     return redirect(
        #         'spider_base:ucontent-list', kwargs={
        #            "token": kwargs.get("source", self.usercomponent).token
        #        }
        #    )
        return ret

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update Travel Protection"))
        # redirect, so user does not see token, needs msg foo for sending token
        # token = self.associated.token
        ret = super().access_update(**kwargs)
        # if (
        #     token != self.associated.token and
        #     TravelProtection.objects.get_active().filter(pk=self.pk).exists()
        # ):
        #     return redirect(
        #         'spider_base:ucontent-list', kwargs={
        #             "token": kwargs.get("source", self.usercomponent).token
        #        }
        #    )
        return ret

    def get_info(self):
        # TODO: cleanup this hack
        # information are directly saved in info, protect field against
        # rewrites
        if self._prepared_info is None:
            return self.associated.info
        return "{}{}".format(
            super().get_info(unlisted=True),
            self._prepared_info
        )

    def save(self, *args, **kwargs):
        ret = super().save(*args, **kwargs)
        self._prepared_info = None
        return ret



__all__ = (
    "get_settings_func", "extract_app_dicts"
)

import inspect
from functools import lru_cache
from importlib import import_module

from django.conf import settings

# for not spamming sets
_empty_set = frozenset()
# for not spamming dicts
_empty_dict = dict()


@lru_cache(maxsize=None)
def get_settings_func(*args, default=None, exclude=frozenset()):
    if default is None:
        names, default = args[:-1], args[-1]
    filterpath = None
    for name in names:
        if getattr(settings, name, None) is not None:
            filterpath = getattr(settings, name)
            break
    if filterpath is None:
        filterpath = default
    # note: cannot distinguish between False and 0 and True and 1
    if filterpath in exclude:
        raise ValueError("invalid setting")
    if callable(filterpath):
        return filterpath
    elif isinstance(filterpath, bool):
        return lambda *args, **kwargs: filterpath
    else:
        module, name = filterpath.rsplit(".", 1)
        return getattr(import_module(module), name)


def extract_app_dicts(app, name, fieldname=None):
    """

    """
    ndic = {}
    for (key, value) in getattr(app, name, _empty_dict).items():
        if inspect.isclass(value):
            if "{}.{}".format(
                value.__module__, value.__qualname__
            ) not in getattr(
                settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
            ):
                if fieldname:
                    setattr(value, fieldname, key)
                ndic[key] = value
        else:
            # set None if Module is not always available
            if value and value not in getattr(
                settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
            ):
                module, name = value.rsplit(".", 1)
                value = getattr(import_module(module, app.name), name)
                if fieldname:
                    setattr(value, fieldname, key)
                ndic[key] = value
    return ndic



__all__ = [
    "rate_limit_default", "allow_all_filter",
    "embed_file_default", "has_admin_permission",
    "LimitedTemporaryFileUploadHandler", "validate_url_default",
    "get_quota", "clean_verifier",
    "clean_verifier_url", "clean_spider_inline"
]

import base64
import logging
import math
import os
import random
import time
from urllib.parse import urlsplit

import requests
from rdflib import XSD, Literal

import ratelimit
from django.conf import settings
from django.core.files.uploadhandler import (
    StopFutureHandlers, StopUpload, TemporaryFileUploadHandler
)
from django.http import Http404
from django.http.request import validate_host
from django.shortcuts import redirect
from django.test import Client
from django.urls import reverse
from django.views.decorators.cache import never_cache
from spkcspider.constants import spkcgraph

from .conf import get_anchor_domain, get_anchor_scheme, get_requests_params
from .signals import failed_guess

# seed with real random
_nonexhaustRandom = random.Random(os.urandom(30))


def rate_limit_default(request, view):
    group = getattr(view, "rate_limit_group", None)
    if group:
        ratelimit.get_ratelimit(
            request=request, group=group, key="user_or_ip",
            inc=True, rate=(math.inf, 3600)
        )
    results = failed_guess.send_robust(sender=view, request=request)
    for (receiver, result) in results:
        if isinstance(result, Exception):
            logging.error(
                "%s failed", receiver, exc_info=result
            )
    # with 0.4% chance reseed
    if _nonexhaustRandom.randint(0, 249) == 0:
        _nonexhaustRandom.seed(os.urandom(10))
    time.sleep(_nonexhaustRandom.random()/2)
    raise Http404()


def allow_all_filter(*args, **kwargs):
    return True


def get_quota(user, quota_type):
    return getattr(user, "quota_{}".format(quota_type), None)


def clean_verifier_url(url):
    if "://" not in url:
        return False
    return True


_default_allowed_hosts = ['localhost', '127.0.0.1', '[::1]']


def clean_spider_inline(url):
    allowed_hosts = settings.ALLOWED_HOSTS
    if settings.DEBUG and not allowed_hosts:
        allowed_hosts = _default_allowed_hosts
    if validate_host(url, allowed_hosts):
        inline_domain = urlsplit(url)
        return inline_domain.netloc or inline_domain.path.split("/", 1)[0]
    return None


def clean_verifier(request, view):
    if not request.auth_token or not request.auth_token.referrer:
        return False
    url = request.auth_token.referrer.url
    if request.method == "POST":
        url = request.POST.get("url", "")
        if not url.startswith(request.auth_token.referrer.url):
            return False
    if "://" not in url:
        return False
    if request.method == "GET":
        # don't spam verifier
        return True
    params, inline_domain = get_requests_params(url)
    if inline_domain:
        response = Client().head(
            url, follow=True, secure=True, Connection="close",
            SERVER_NAME=inline_domain
        )
        if response.status_code >= 400:
            return False
    else:
        try:
            with requests.head(
                url, stream=True, **params,
                headers={"Connection": "close"}
            ) as resp:
                resp.close()
                resp.raise_for_status()
        except Exception:
            return False
    return True


class LimitedTemporaryFileUploadHandler(TemporaryFileUploadHandler):
    activated = False

    def handle_raw_input(
        self, input_data, META, content_length, boundary, encoding=None
    ):
        """
        Use the content_length to signal whether or not this handler should be
        used.
        """
        # disable upload if too big
        self.activated = self.check_allowed_size(content_length)

    def new_file(self, *args, **kwargs):
        if not self.activated:
            raise StopFutureHandlers()
        return super().new_file(*args, **kwargs)

    def receive_data_chunk(self, raw_data, start):
        if not self.activated:
            raise StopUpload(True)
        return super().receive_data_chunk(raw_data, start)

    def file_complete(self, file_size):
        """Return a file object if this handler is activated."""
        if not self.activated:
            return
        return super().file_complete(file_size)

    def check_allowed_size(self, content_length):
        if self.request.user.is_staff:
            max_length = getattr(
                settings, "SPIDER_MAX_FILE_SIZE_STAFF", None
            )
        else:
            max_length = getattr(
                settings, "SPIDER_MAX_FILE_SIZE", None
            )
        if not max_length:
            return True

        # superuser can upload as much as he wants
        if self.request.user.is_superuser:
            return True

        return content_length <= max_length


def validate_url_default(url, view=None):
    url = urlsplit(url)
    if url.scheme == "https":
        return True
    elif url.scheme == "http":
        # MAKE SURE that you control your dns
        if url.netloc.endswith(".onion"):
            return True
        elif (
            get_anchor_domain() == url.netloc and
            get_anchor_scheme() == "http"
        ):
            return True
        elif settings.DEBUG:
            return True
    return False


def embed_file_default(name, value, content, context):

    override = (
        (
            context["request"].user.is_superuser or
            context["request"].user.is_staff
        ) and context["request"].GET.get("embed_big", "") == "true"
    )
    if (
        value.size < getattr(settings, "MAX_EMBED_SIZE", 4000000) or
        override
    ):
        return Literal(
            base64.b64encode(value.read()),
            datatype=XSD.base64Binary,
            normalize=False
        )
    elif (
        context["scope"] == "export" or
        getattr(settings, "FILE_DIRECT_DOWNLOAD", False)
    ):
        # link always direct to files in exports
        url = value.url
        if "://" not in getattr(settings, "MEDIA_URL", ""):
            url = "{}{}".format(context["hostpart"], url)
        return Literal(
            url,
            datatype=spkcgraph["hashableURI"],
        )
    else:
        # only file filet has files yet
        url = content.associated.get_absolute_url("download")
        url = "{}{}?{}".format(
            context["hostpart"],
            url, context["context"]["sanitized_GET"]
        )
        return Literal(
            url,
            datatype=spkcgraph["hashableURI"],
        )


def has_admin_permission(obj, request):
    # allow only non travel protected user with superuser and staff permissions
    if not request.user.is_active or not request.user.is_staff:
        return False
    # user is authenticate now
    assert request.user.is_authenticated
    if not request.session["is_travel_protected"]:
        from .models import AssignedContent
        t = AssignedContent.travel.get_active_for_request(
            request
        ).exists()
        # auto activates on trying to access admin but not deactivate
        if t:
            request.session["is_travel_protected"] = t
    if request.session["is_travel_protected"]:
        return False
    return True


@never_cache
def admin_login(self, request, extra_context=None):
    """
    Display the login form for the given HttpRequest.
    """
    if request.method == 'GET' and self.has_permission(request):
        # Already logged-in, redirect to admin index
        index_path = reverse('admin:index', current_app=self.name)
        return redirect(index_path)
    else:
        loginpath = getattr(
            settings,
            "LOGIN_URL",
            reverse("auth:login")
        )
        return redirect(loginpath)



__all__ = ["FieldRegistry", "LayoutRegistry", "fields", "layouts"]

from spkcspider.apps.spider.registry import Registry


class FieldRegistry(Registry):
    attr_path = "spider_fields_path"

    def __setitem__(self, key, value):
        assert value, "invalid value type %s" % type(value)
        super().__setitem__(key, value)


class LayoutRegistry(Registry):
    attr_path = "spider_layouts_path"

    def __setitem__(self, key, value):
        assert value, "invalid value type %s" % type(value)
        super().__setitem__(key, value)

    def initialize(self):
        from django.apps import apps
        TagLayout = apps.get_model("spider_tags", "TagLayout")
        # iterate over unionized layouts
        for name, layout_dic in self.items():
            tag_layout = TagLayout.objects.get_or_create(
                defaults=layout_dic, name=name
            )[0]
            has_changed = False
            # check attributes of model
            for layout_attr, value in layout_dic.items():
                if getattr(tag_layout, layout_attr) != value:
                    # layout change detected, update and mark for saving
                    setattr(tag_layout, layout_attr, value)
                    has_changed = True

            if has_changed:
                tag_layout.save()


fields = FieldRegistry()
layouts = LayoutRegistry()



# Generated by Django 3.0.2 on 2020-01-18 14:47

from django.db import migrations, models
import django.db.models.deletion
import django.db.models.expressions
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0012_auto_20191230_1305'),
    ]

    operations = [
        migrations.CreateModel(
            name='SmartTag',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, default='', max_length=50)),
                ('unique', models.BooleanField(blank=True, default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('data', jsonfield.fields.JSONField(blank=True, default=dict)),
                ('free', models.BooleanField(default=False, editable=False)),
                ('content', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='smarttags', to='spider_base.AssignedContent')),
                ('target', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='smarttag_sources', to='spider_base.AssignedContent')),
            ],
        ),
        migrations.AddConstraint(
            model_name='smarttag',
            constraint=models.UniqueConstraint(condition=models.Q(unique=True), fields=('content', 'target', 'name'), name='smarttag_target_unique'),
        ),
        migrations.AddConstraint(
            model_name='smarttag',
            constraint=models.UniqueConstraint(condition=models.Q(('target__isnull', True), ('unique', True)), fields=('content', 'name'), name='smarttag_notarget_unique'),
        ),
        migrations.AddConstraint(
            model_name='smarttag',
            constraint=models.CheckConstraint(check=models.Q(_negated=True, content=django.db.models.expressions.F('target')), name='smarttag_content_not_target'),
        ),
    ]



"""
Does require celery only for async methods
"""
__all__ = {
    "validate", "valid_wait_states", "verify_download_size",
    "async_validate", "verify_tag", "async_verify_tag"
}

import io
import logging
import tempfile

import requests
from rdflib import Graph, Literal, URIRef
from rdflib.namespace import XSD

from django.conf import settings
from django.core import exceptions
from django.core.files import File
from django.test import Client
from django.utils.translation import gettext as _
from spkcspider import celery_app
from spkcspider.constants.rdf import spkcgraph
from spkcspider.utils.settings import get_settings_func
from spkcspider.utils.urls import merge_get_url

from .conf import get_requests_params
# uses specialized get_hashob from verifier (can be further customized)
from .functions import get_anchor_domain, get_hashob
from .models import DataVerificationTag, VerifySourceObject

logger = logging.getLogger(__name__)


BUFFER_SIZE = 65536  # read in 64kb chunks

valid_wait_states = {
    "RETRIEVING", "HASHING", "STARTED"
}


def verify_download_size(length, current_size=0):
    if not length or not length.isdigit():
        return False
    length = int(length)
    if settings.VERIFIER_MAX_SIZE_ACCEPTED < length:
        return False
    return True


def retrieve_object(obj, current_size, graph=None, session=None):
    # without graph return hash
    ret = None
    try:
        if not isinstance(obj, str):
            if graph is not None:
                graph.parse(obj, format="turtle")
            else:
                h = get_hashob()
                h.update(XSD.base64Binary.encode("utf8"))
                for chunk in iter(lambda: obj.read(BUFFER_SIZE), b''):
                    h.update(chunk)
                return h.finalize()

            return
        # obj is url
        params, inline_domain = get_requests_params(obj)
        if inline_domain:
            resp = Client().get(obj, SERVER_NAME=inline_domain)
            if resp.status_code != 200:
                raise exceptions.ValidationError(
                    _("Retrieval failed: %(reason)s"),
                    params={"reason": resp.reason},
                    code="error_code:{}".format(resp.status_code)
                )

            c_length = resp.get("content-length", None)
            if (
                c_length is None or
                not verify_download_size(c_length, current_size[0])
            ):
                resp.close()
                raise exceptions.ValidationError(
                    _("Content too big or size unset: %(size)s"),
                    params={"size": c_length},
                    code="invalid_size"
                )

            c_length = int(c_length)
            current_size[0] += c_length
            if graph is not None:
                graph.parse(getattr(
                    resp, "streaming_content", io.BytesIO(resp.content)
                ), format="turtle")
            else:
                h = get_hashob()
                h.update(XSD.base64Binary.encode("utf8"))
                for chunk in resp:
                    h.update(chunk)
                ret = h.finalize()
            resp.close()
        else:
            if not session:
                session = requests
            try:
                with session.get(
                    obj, stream=True, **params
                ) as resp:
                    if resp.status_code != 200:
                        raise exceptions.ValidationError(
                            _("Retrieval failed: %(reason)s"),
                            params={"reason": resp.reason},
                            code="error_code:{}".format(resp.status_code)
                        )

                    c_length = resp.headers.get("content-length", None)
                    if (
                        c_length is None or
                        not verify_download_size(c_length, current_size[0])
                    ):
                        raise exceptions.ValidationError(
                            _("Content too big or size unset: %(size)s"),
                            params={"size": c_length},
                            code="invalid_size"
                        )
                    c_length = int(c_length)
                    current_size[0] += c_length
                    if graph is not None:
                        graph.parse(resp.raw, format="turtle")
                    else:
                        h = get_hashob()
                        h.update(XSD.base64Binary.encode("utf8"))
                        for chunk in resp.iter_content(BUFFER_SIZE):
                            h.update(chunk)
                        ret = h.finalize()

            except requests.exceptions.Timeout:
                raise exceptions.ValidationError(
                    _('url timed out: %(url)s'),
                    params={"url": obj},
                    code="timeout_url"
                )
            except requests.exceptions.ConnectionError:
                raise exceptions.ValidationError(
                    _('invalid url: %(url)s'),
                    params={"url": obj},
                    code="invalid_url"
                )
    except Exception:
        logger.error("Parsing graph failed", exc_info=settings.DEBUG)
        raise exceptions.ValidationError(
            _('Invalid graph fromat'),
            code="invalid_format"
        )
    return ret


def validate(ob, hostpart, task=None):
    dvfile = None
    source = None
    g = Graph()
    g.namespace_manager.bind("spkc", spkcgraph, replace=True)
    with requests.session() as session:
        if isinstance(ob, tuple):
            current_size = ob[1]
            with open(ob[0], "rb") as f:
                retrieve_object(f, [current_size], graph=g, session=session)
        else:
            current_size = 0

            source = VerifySourceObject.objects.get(
                id=ob
            )
            url = source.get_url()
            retrieve_object(url, [current_size], graph=g, session=session)

        tmp = list(g.query(
            """
                SELECT ?base ?scope ?pages ?view
                WHERE {
                    ?base spkc:scope ?scope ;
                          spkc:pages.num_pages ?pages ;
                          spkc:pages.current_page ?current_page ;
                          spkc:action:view ?view .
                }
            """,
            initNs={"spkc": spkcgraph},
            initBindings={
                "current_page": Literal(1, datatype=XSD.positiveInteger)
            }
        ))

        if len(tmp) != 1:
            raise exceptions.ValidationError(
                _('Invalid graph'),
                code="invalid_graph"
            )
        tmp = tmp[0]
        start = tmp.base
        # scope = tmp.scope.toPython()
        view_url = tmp.view.toPython()
        pages = tmp.pages.toPython()

        if task:
            task.update_state(
                state='RETRIEVING',
                meta={
                    'page': 1,
                    'num_pages': pages
                }
            )
        if isinstance(ob, tuple):
            split = view_url.split("?", 1)
            source = VerifySourceObject.objects.update_or_create(
                url=split[0], defaults={"get_params": split[1]}
            )

        # retrieve further pages
        for page in range(2, pages+1):
            url = merge_get_url(
                source.get_url(), raw="embed", page=str(page)
            )
            retrieve_object(url, [current_size], graph=g, session=session)

            if task:
                task.update_state(
                    state='RETRIEVING',
                    meta={
                        'page': page,
                        'num_pages': pages
                    }
                )

        # check and clean graph
        data_type = get_settings_func(
            "VERIFIER_CLEAN_GRAPH",
            "spkcspider.apps.verifier.functions.clean_graph"
        )(g, start, source, hostpart)
        if not data_type:
            raise exceptions.ValidationError(
                _('Invalid graph (Verification failed)'),
                code="graph_failed"
            )
        g.remove((None, spkcgraph["csrftoken"], None))

        hashable_nodes = g.query(
            """
                SELECT DISTINCT ?base ?type ?name ?value
                WHERE {
                    ?base spkc:type ?type ;
                          spkc:properties ?prop .
                    ?prop spkc:hashable "true"^^xsd:boolean ;
                          spkc:name ?name ;
                          spkc:value ?value .
                    OPTIONAL {
                        ?value spkc:properties ?prop2 ;
                               spkc:name "info"^^xsd:string ;
                               spkc:value ?info .
                    } .
                }
            """,
            initNs={"spkc": spkcgraph}
        )

        if task:
            task.update_state(
                state='HASHING',
                meta={
                    'hashable_nodes_checked': 0
                }
            )
        # make sure triples are linked to start
        # (user can provide arbitary data)
        g.remove((start, spkcgraph["hashed"], None))
        # MAYBE: think about logic for incoperating hashes
        g.remove((start, spkcgraph["hash"], None))
        nodes = {}
        resources_with_hash = {}
        for count, val in enumerate(hashable_nodes, start=1):
            if isinstance(val.value, URIRef):
                assert(val.info)
                h = get_hashob()
                # should always hash with xsd.string
                h.update(XSD.string.encode("utf8"))
                # don't strip id, as some contents seperate only by id
                h.update(str(val.info).encode("utf8"))
                _hash = h.finalize()
            elif val.value.datatype == spkcgraph["hashableURI"]:
                _hash = resources_with_hash.get(val.value.value)
                if not _hash:
                    url = merge_get_url(val.value.value, raw="embed")
                    if not get_settings_func(
                        "SPIDER_URL_VALIDATOR",
                        "spkcspider.apps.spider.functions.validate_url_default"
                    )(url):
                        raise exceptions.ValidationError(
                            _('invalid url: %(url)s'),
                            params={"url": url},
                            code="invalid_url"
                        )
                    _hash = retrieve_object(
                        url, [current_size], session=session
                    )
                    # do not use add as it could be corrupted by user
                    # (user can provide arbitary data)
                    _uri = URIRef(val.value.value)
                    g.set((
                        _uri,
                        spkcgraph["hash"],
                        Literal(_hash.hex())
                    ))
            else:
                h = get_hashob()
                if val.value.datatype == XSD.base64Binary:
                    h.update(val.value.datatype.encode("utf8"))
                    h.update(val.value.toPython())
                elif val.value.datatype:
                    h.update(val.value.datatype.encode("utf8"))
                    h.update(val.value.encode("utf8"))
                else:
                    h.update(XSD.string.encode("utf8"))
                    h.update(val.value.encode("utf8"))
                _hash = h.finalize()
            h = get_hashob()
            h.update(val.name.toPython().encode("utf8"))
            h.update(_hash)
            base = str(val.base)
            nodes.setdefault(base, ([], val.type))
            nodes[base][0].append(h.finalize())
            if task:
                task.update_state(
                    state='HASHING',
                    meta={
                        'hashable_nodes_checked': count
                    }
                )
    if task:
        task.update_state(
            state='HASHING',
            meta={
                'hashable_nodes_checked': "all"
            }
        )

    # first sort hashes per node and create hash over sorted hashes
    # de-duplicate super-hashes (means: nodes are identical)
    hashes = set()
    for val, _type in nodes.values():
        h = get_hashob()
        for _hob in sorted(val):
            h.update(_hob)
        h.update(_type.encode("utf8"))
        hashes.add(h.finalize())

    # then create hash over sorted de-duplicated node hashes
    h = get_hashob()
    for i in sorted(hashes):
        h.update(i)
    # do not use add as it could be corrupted by user
    # (user can provide arbitary data)
    digest = h.finalize().hex()
    g.set((
        start,
        spkcgraph["hash"],
        Literal(digest)
    ))

    with tempfile.NamedTemporaryFile(delete=True) as dvfile:
        # save in temporary file
        g.serialize(
            dvfile, format="turtle"
        )

        result, created = DataVerificationTag.objects.get_or_create(
            defaults={
                "dvfile": File(dvfile),
                "source": source,
                "data_type": data_type
            },
            hash=digest
        )
    update_fields = set()
    # and source, cannot remove source without replacement
    if not created and source and source != result.source:
        result.source = source
        update_fields.add("source")
    if data_type != result.data_type:
        result.data_type = data_type
        update_fields.add("data_type")
    result.save(update_fields=update_fields)
    verify_tag(result, task=task, ffrom="validate")
    if task:
        task.update_state(
            state='SUCCESS'
        )
    return result


if celery_app:
    @celery_app.task(bind=True, name='async validation')
    def async_validate(self, ob, hostpart):
        ret = validate(ob, hostpart, self)
        return ret.get_absolute_url()
else:
    def async_validate(self, ob, hostpart):
        raise Exception("no celery installed")


def verify_tag(tag, hostpart=None, ffrom="sync_call", task=None):
    """ for auto validation or hooks"""
    if not hostpart:
        hostpart = get_anchor_domain()
    if task:
        task.update_state(
            state='VERIFY'
        )

    if get_settings_func(
        "VERIFIER_TAG_VERIFIER",
        "spkcspider.apps.verifier.functions.verify_tag_default"
    )(tag, hostpart, ffrom):
        try:
            tag.callback(hostpart)
        except exceptions.ValidationError:
            logger.exception("Error while calling back")
    if task:
        task.update_state(
            state='SUCCESS'
        )


if celery_app:
    @celery_app.task(bind=True, name='async verification', ignore_results=True)
    def async_verify_tag(self, tagid, hostpart=None, ffrom="async_call"):
        verify_tag(
            tag=DataVerificationTag.objects.get(id=tagid),
            hostpart=hostpart, task=self, ffrom=ffrom
        )
else:
    def async_verify_tag(self, tagid, hostpart=None, ffrom="async_call"):
        raise Exception("no celery installed")



__all__ = [
    "LinkForm", "TravelProtectionForm", "TravelProtectionManagementForm"
]

import json
from base64 import b64encode
from datetime import timedelta as td

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from django import forms
from django.conf import settings
from django.contrib.auth import authenticate
from django.db import models
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.utils.translation import gettext_lazy as _

from spkcspider.constants import (
    TravelProtectionType, VariantType, dangerous_login_choices,
    travel_scrypt_params
)

from ..fields import ContentMultipleChoiceField, MultipleOpenChoiceField
from ..models import AssignedContent, AttachedTimespan, UserComponent
from ..queryfilters import loggedin_active_tprotections_q
from ..widgets import (
    ListWidget, OpenChoiceWidget, SelectizeWidget
)
from ._messages import travel_protection as _travel_protection
from ._messages import time_help_text
from .base import DataContentForm

_extra = '' if settings.DEBUG else '.min'


class LinkForm(DataContentForm):
    content = forms.ModelChoiceField(
        queryset=AssignedContent.objects.filter(
            strength__lte=10
        )
    )
    push = forms.BooleanField(
        required=False,
        initial=False,
        help_text=_("Improve ranking of this Link.")
    )

    free_fields = {"push": False}

    def __init__(self, uc, request, **kwargs):
        super().__init__(**kwargs)
        # if self.instance.associated:
        #     if "\x1eanchor\x1e" in self.instance.associated:
        #         self.fields["content"].disabled = True
        if self.instance.id:
            self.initial["content"] = \
                self.instance.associated.attached_to_content
        q = self.fields["content"].queryset
        travel = AssignedContent.travel.get_active_for_request(request)
        travel = travel.filter(loggedin_active_tprotections_q)
        self.fields["content"].queryset = q.filter(
            strength__lte=uc.strength
        ).exclude(
            models.Q(usercomponent__travel_protected__in=travel) |
            models.Q(travel_protected__in=travel)
        )
        # component auth should limit links to visible content
        # read access outside from component elsewise possible
        if request.user != uc.user and not request.is_staff:
            q = self.fields["content"].queryset
            self.fields["content"].queryset = q.filter(
                models.Q(usercomponent=uc) |
                models.Q(referenced_by__usercomponent=uc)
            )

    def save(self, commit=True):
        self.instance.associated.attached_to_content = \
            self.cleaned_data["content"]
        return super().save(commit)


class TravelProtectionManagementForm(DataContentForm):

    def clean(self):
        super().clean()
        pwset = set(self.instance.associated.getlist("pwhash", 20))
        self.instance._prepared_info = "".join(
            map(lambda x: "pwhash={}\x1e".format(x), pwset)
        )
        if self.instance.associated.getflag("anonymous_deactivation"):
            self.instance._prepared_info = \
                "{}anonymous_deactivation\x1e".format(
                    self.instance._prepared_info
                )
        if self.instance.associated.getflag("anonymous_trigger"):
            self.instance._prepared_info = \
                "{}anonymous_trigger\x1e".format(
                    self.instance._prepared_info
                )

        return self.cleaned_data


class TravelProtectionForm(DataContentForm):
    request = None
    pwset = None
    # self_protection = forms.ChoiceField(
    #    label=_("Self-protection"), help_text=_(_self_protection),
    #     initial="None", choices=PROTECTION_CHOICES
    # )

    active = forms.BooleanField(initial=True, required=False)
    timeplans = MultipleOpenChoiceField(
        widget=ListWidget(
            items=[
                {
                    "name": "start",
                    "label": _("Start"),
                    "format_type": "datetime-local",
                    "options": {
                        "flatpickr": {
                            "inline": True,
                            "time_24hr": True,
                            "": "Z"
                        }
                    }
                },
                {
                    "name": "stop",
                    "label": _("Stop"),
                    "format_type": "datetime-local",
                    "options": {
                        "flatpickr": {
                            "inline": True,
                            "time_24hr": True,
                            "": "Z"
                        }
                    }
                }
            ], item_label=_("Timeplan")
        ), required=False,
        help_text=time_help_text
    )
    trigger_pws = MultipleOpenChoiceField(
        label=_("Trigger Passwords"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )
    overwrite_pws = forms.BooleanField(
        required=False, initial=False,
        help_text=_(
            "If set, the set of passwords is overwritten or "
            "removed (if no passwords are set)"
        )
    )
    anonymous_deactivation = forms.BooleanField(
        required=False, initial=False,
        help_text=_(
            "Can deactivate protection without login"
        )
    )

    protect_contents = ContentMultipleChoiceField(
        queryset=AssignedContent.objects.all(), required=False,
        to_field_name="id",
        widget=SelectizeWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )
    protect_components = forms.ModelMultipleChoiceField(
        queryset=UserComponent.objects.all(), required=False,
        to_field_name="name",
        widget=SelectizeWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    travel_protection_type = forms.ChoiceField(
        choices=TravelProtectionType.as_choices(),
        initial=TravelProtectionType.hide,
        help_text=_travel_protection
    )

    master_pw = forms.CharField(
        label=_("Master Login Password"),
        widget=forms.PasswordInput(render_value=True), required=True,
        help_text=_(
            "Enter Password used for the user account"
        )
    )

    quota_fields = {"trigger_pws": list}

    class Media:
        css = {
            "all": [
                "node_modules/flatpickr/dist/flatpickr%s.css" % _extra
            ]
        }
        js = [
            'node_modules/flatpickr/dist/flatpickr%s.js' % _extra
        ]

    def _filter_dangerousprotection(self, x):
        # if it is already selected, still allow selection
        #   elsewise confusing logic and errors
        if (
            x[0] in dangerous_login_choices and
            x[0] != self.initial["travel_protection_type"]
        ):
            return False
        return True

    @staticmethod
    def _filter_selfprotection(x):
        if x[0] == TravelProtectionType.disable:
            return False
        return True

    def __init__(self, request, **kwargs):
        super().__init__(**kwargs)
        self.request = request
        self.pwset = set()
        associated = self.instance.associated
        self.initial["active"] = associated.getflag("active")
        self.initial["anonymous_deactivation"] = \
            associated.getflag("anonymous_deactivation")

        self.fields["timeplans"].help_text = str(
            self.fields["timeplans"].help_text
        ).format(timezone=timezone.get_current_timezone_name())
        if not getattr(self.instance, "id"):
            now = timezone.now()
            self.initial["travel_protection_type"] = TravelProtectionType.hide
            self.initial["timeplans"] = \
                [
                    json.dumps({
                        "start": (now + td(hours=3)).isoformat(),
                        "stop": (now + td(days=7)).isoformat()
                    })
                ]
        else:
            self.initial["travel_protection_type"] = \
                associated.getlist("travel_protection_type", 1)[0]
            self.initial["timeplans"] = \
                [
                    json.dumps({
                        "start": x.start.isoformat() if x.start else None,
                        "stop": x.stop.isoformat() if x.stop else None
                    }) for x in associated.attachedtimespans.filter(
                        name="active"
                    )
                ]
        if not self.request.user.has_perm(
            "spider_base.use_dangerous_travelprotections"
        ):
            self.fields["travel_protection_type"].choices = \
                filter(
                    self._filter_dangerousprotection,
                    self.fields["travel_protection_type"].choices
                )
        if associated.ctype.name == "SelfProtection":
            self.fields["travel_protection_type"].choices = \
                filter(
                    self._filter_selfprotection,
                    self.fields["travel_protection_type"].choices
                )
            del self.fields["timeplans"]
        if not getattr(self.instance, "id") and not self.data:
            self.initial["active"] = True

        travel = AssignedContent.travel.get_active_for_request(
            request
        ).filter(loggedin_active_tprotections_q)
        selfid = self.instance.id or -1
        q_component = models.Q(
            user=request.user
        ) & (
            ~models.Q(travel_protected__in=travel) |
            models.Q(travel_protected__id=selfid)
        )

        q_content = models.Q(usercomponent__user=request.user) & (
            ~(
                models.Q(usercomponent__travel_protected__in=travel) |
                models.Q(travel_protected__in=travel) |
                # don't allow detectable contents
                models.Q(info__contains="\x1eanchor\x1e") |
                models.Q(info__contains="\x1eprimary\x1e") |
                # contents also appearing as features are easily detectable
                models.Q(
                    ctype__ctype__contains=VariantType.feature_connect
                )
            ) |
            (
                models.Q(usercomponent__travel_protected__id=selfid) |
                models.Q(travel_protected__id=selfid)
            )
        )
        if getattr(self.instance, "id", None):
            # exlude own content
            q_content &= ~models.Q(pk=associated.pk)
            self.initial["anonymous_deactivation"] = \
                associated.getflag("anonymous_deactivation")
            if associated.getlist("pwhash", 1):
                self.fields["trigger_pws"].help_text = _(
                    "<span style='color:red'>Trigger passwords already set."
                    "</span>"
                )
            else:
                self.fields["trigger_pws"].help_text = _(
                    "No trigger passwords set."
                )
                del self.fields["overwrite_pws"]
        if (
            not associated.getlist("pwhash", 1) and
            associated.ctype.name == "SelfProtection"
        ):
            self.fields["trigger_pws"].required = True
        # use q for filtering (including own)
        self.fields["protect_contents"].queryset = \
            self.fields["protect_contents"].queryset.filter(
                q_content
            ).distinct().order_by(
                "name"
            )

        self.fields["protect_components"].queryset = \
            self.fields["protect_components"].queryset.filter(
                q_component
            ).distinct().order_by(
                "name"
            )

    def _clean_timeplan(self, ob):
        if isinstance(ob, str):
            ob = json.loads(ob)
        start = ob.get("start", None)
        if isinstance(start, str):
            start = parse_datetime(start)
        stop = ob.get("stop", None)
        if isinstance(stop, str):
            stop = parse_datetime(stop)
        return AttachedTimespan(
            unique=False,
            name="active",
            start=start,
            stop=stop,
            content=self.instance.associated
        )

    def clean_timeplans(self):
        return list(map(self._clean_timeplan, self.cleaned_data["timeplans"]))

    def clean(self):
        super().clean()
        self.initial["master_pw"] = self.cleaned_data.pop("master_pw", None)
        if not authenticate(
            self.request,
            username=self.instance.associated.usercomponent.username,
            password=self.initial.get("master_pw", None),
            nospider=True
        ):
            self.add_error(
                "master_pw",
                forms.ValidationError(
                    _("Invalid Password"),
                    code="invalid_password"
                )
            )
        if "trigger_pws" in self.cleaned_data:
            pwset = set(map(
                lambda x: b64encode(Scrypt(
                    salt=settings.SECRET_KEY.encode("utf-8"),
                    backend=default_backend(),
                    **travel_scrypt_params
                ).derive(x[:128].encode("utf-8"))).decode("ascii"),
                self.cleaned_data["trigger_pws"]
            ))
            if (
                self.instance.associated.ctype.name == "SelfProtection" and
                self.cleaned_data.get("overwrite_pws") and
                len(pwset) == 0
            ):
                # = passwords are completely empty
                self.add_error("trigger_pws", forms.ValidationError(
                    _(
                        "Empty trigger passwords not allowed "
                        "(for Self-Protection)"
                    )
                ))
            if not self.cleaned_data.get("overwrite_pws") and len(pwset) < 20:
                pwset.update(self.instance.associated.getlist(
                    "pwhash", 20-len(pwset)
                ))
            self.pwset = pwset
        else:
            self.pwset = set(self.instance.associated.getlist("pwhash", 20))
        for i in self.cleaned_data["timeplans"]:
            try:
                i.clean()
            except forms.ValidationError as exc:
                self.add_error("timeplans", exc)
        try:
            self.cleaned_data["protect_contents"] = \
                self.cleaned_data["protect_contents"].exclude(
                    usercomponent__in=self.cleaned_data["protect_components"]
                )
        except KeyError:
            pass

        return self.cleaned_data

    def is_valid(self):
        isvalid = super().is_valid()
        if not getattr(self, "cleaned_data", None):
            return False
        return isvalid

    def get_prepared_attachements(self):
        ret = {
            "attachedtimespans": self.cleaned_data["timeplans"]
        }
        if self.cleaned_data["travel_protection_type"] not in {
            TravelProtectionType.wipe_user.value,
            TravelProtectionType.trigger_disable_user.value
        }:
            ret["protect_contents"] = \
                [
                    self.instance.associated,
                    *self.cleaned_data["protect_contents"]
                ]
            ret["protect_components"] = \
                list(self.cleaned_data["protect_components"])
        return ret

    def save(self, commit=False):
        self.instance.free_data["is_travel_protected"] = \
            self.request.session.get("is_travel_protected", False)
        # activates _prepared_info
        self.instance._prepared_info = ""
        if self.cleaned_data.get("active"):
            self.instance._prepared_info = \
                "{}active\x1e".format(
                    self.instance._prepared_info
                )
        self.instance._prepared_info = \
            "{}travel_protection_type={}\x1e".format(
                self.instance._prepared_info,
                self.cleaned_data["travel_protection_type"]
            )
        if self.cleaned_data.get("anonymous_deactivation"):
            self.instance._prepared_info = \
                "{}anonymous_deactivation\x1e".format(
                    self.instance._prepared_info
                )
        if self.cleaned_data.get("anonymous_trigger"):
            self.instance._prepared_info = \
                "{}anonymous_trigger\x1e".format(
                    self.instance._prepared_info
                )
        self.instance._prepared_info += "".join(
            map(lambda x: "pwhash={}\x1e".format(x), self.pwset)
        )
        return super().save(commit)




import logging
from itertools import chain

from rdflib import URIRef

from django.core.exceptions import ValidationError
from django.db import models
from django.http import HttpResponse
from django.urls import reverse
from django.utils.html import escape
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext
from django.views.decorators.csrf import csrf_exempt
from jsonfield import JSONField
from spkcspider.apps.spider.abstract_models import BaseContent
from spkcspider.apps.spider.models import DataContent
from spkcspider.apps.spider import registry
from spkcspider.constants import ActionUrl, VariantType
from spkcspider.utils.settings import get_settings_func
from spkcspider.utils.fields import add_by_field

from .generators import generate_form

logger = logging.getLogger(__name__)

# Create your models here.


class TagLayout(models.Model):
    name = models.SlugField(max_length=255, null=False)
    # shall it be unique for a component?
    unique = models.BooleanField(default=False, blank=True)
    layout = JSONField(
        default=list,
        help_text=_("Field list in JSON format")
    )
    default_verifiers = JSONField(default=list, blank=True)
    usertag = models.OneToOneField(
        "spider_base.AssignedContent", on_delete=models.CASCADE,
        related_name="+", null=True, blank=True
    )
    description = models.TextField(default="", blank=True)

    class Meta(object):
        unique_together = [
            ("name", "usertag")
        ]

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Tag Layout")

    def get_description(self):
        return self.description

    def full_clean(self, **kwargs):
        # checked with clean
        kwargs.setdefault("exclude", []).append("usertag")
        return super().full_clean(**kwargs)

    def clean(self):
        if self.usertag:
            if TagLayout.objects.filter(usertag=None, name=self.name).exists():
                raise ValidationError(
                    _("Layout exists already"),
                    code="unique"
                )
            self.usertag.full_clean(exclude=["layout"])

    def get_form(self):
        return generate_form("LayoutForm", self.layout)

    def __repr__(self):
        if self.usertag:
            return "<TagLayout: %s:%s>" % (
                self.name, self.usertag.associated.usercomponent.username
            )
        return "<TagLayout: %s>" % self.name

    def __str__(self):
        if self.usertag:
            return "TagLayout: %s:%s" % (
                self.name,
                self.usertag.associated.usercomponent.username
            )
        return "TagLayout: %s" % self.name


@add_by_field(registry.contents, "_meta.model_name")
class UserTagLayout(DataContent):
    # 10 is required for preventing info leak gadgets via component auth
    appearances = [
        {
            "name": "TagLayout",
            "ctype": VariantType.unique,
            "strength": 10
        }
    ]
    expose_name = False
    expose_description = True

    class Meta:
        proxy = True

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Tag Layout")

    def get_content_name(self):
        return "%s: %s" % (
            self.associated.usercomponent.username,
            self.layout.name,
        )

    def localized_description(self):
        """ localize and perform other transforms before rendering to user """
        return gettext(self.associated.description)

    def get_size(self, prepared_attachements=None):
        s = super().get_size(prepared_attachements)
        s += len(str(self.layout.default_verifiers))
        s += len(str(self.layout.layout))
        return s

    def get_template_name(self, scope):
        if scope == "view":
            return 'spider_tags/edit_preview_form.html'
        return super().get_template_name(scope)

    def get_strength_link(self):
        # never allow links to this, elsewise with links is an information
        # disclosure possible
        return 11

    def get_info(self):
        return "%slayout=%s\x1e" % (
            super().get_info(),
            self.layout.name
        )

    def get_form(self, scope):
        if scope in {"add", "update", "export"}:
            from .forms import TagLayoutForm
            return TagLayoutForm
        else:
            return self.layout.get_form()

    def access_view(self, **kwargs):
        _ = gettext
        kwargs.setdefault(
            "legend",
            escape(_("Check \"%s\"") % self.__str__())
        )
        # not visible by default
        kwargs.setdefault("confirm", _("Check"))
        # prevent second button
        kwargs.setdefault("inner_form", False)
        return super().access_view(**kwargs)

    def get_form_kwargs(self, **kwargs):
        kwargs["instance"] = self.layout
        ret = super().get_form_kwargs(**kwargs)
        if kwargs["scope"] in {"add", "update", "export"}:
            ret["usertaglayout"] = self
        return ret

    def access(self, context):
        if context["scope"] == "view":
            context["extra_outer_forms"] = ["request_verification_form"]
        return super().access(context)

    @property
    def layout(self):
        tlayout = None
        if getattr(self.associated, "pk", None):
            tlayout = TagLayout.objects.filter(usertag=self.associated).first()
        if not tlayout:
            tlayout = TagLayout(usertag=self.associated)
        return tlayout


@add_by_field(registry.contents, "_meta.model_name")
class SpiderTag(BaseContent):
    # cannot optimize into a DataContent, too many ForeignKeys
    _cached_references = None
    tmp_primary_anchor = None
    appearances = [
        {
            "name": "SpiderTag",
            "strength": 0,
            "ctype": VariantType.domain_mode
        },
        {
            "name": "PushedTag",
            "strength": 2,
            "ctype": VariantType.component_feature + VariantType.domain_mode
        }
    ]
    layout = models.ForeignKey(
        TagLayout, related_name="tags", on_delete=models.PROTECT,
    )
    tagdata = JSONField(default=dict, blank=True)
    verified_by = JSONField(default=list, blank=True)
    updateable_by = models.ManyToManyField(
        "spider_base.ReferrerObject", related_name="tags", blank=True
    )
    primary = models.BooleanField(default=False, blank=True)

    expose_name = False
    expose_description = False

    def __str__(self):
        if not self.id:
            return self.localize_name(self.associated.ctype.name)
        if not self.layout.usertag:
            return "%s: <%s>: %s" % (
                self.localize_name("SpiderTag"),
                self.layout.name,
                self.associated_id
            )
        return "%s: <%s: %s>: %s" % (
            self.localize_name("SpiderTag"),
            self.layout.usertag.usercomponent.username,
            self.layout.name,
            self.associated_id
        )

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        if name == "PushedTag":
            return _("content name", "Allow receiving tags")
        else:
            return _("content name", "Tag")

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl("pushtag", reverse("spider_tags:create-pushtag"))
        ]

    def get_content_description(self):
        return self.layout.get_description()

    def get_content_name(self):
        if not self.layout.usertag:
            return self.layout.name
        return self.layout.usertag.content.get_content_name()

    def get_template_name(self, scope):
        if scope == "update":
            return 'spider_tags/edit_form.html'
        if scope == "push_update":
            return 'spider_base/edit_form.html'
        return super().get_template_name(scope)

    def get_size(self, prepared_attachements=None):
        s = super().get_size(prepared_attachements)
        s += len(str(self.verified_by))
        s += len(str(self.tagdata))
        return s

    def get_strength_link(self):
        return 0

    def get_abilities(self, context):
        _abilities = set()
        if (
            context["request"].auth_token and
            context["request"].auth_token.referrer
        ):
            if get_settings_func(
                "SPIDER_TAG_VERIFIER_VALIDATOR",
                "spkcspider.apps.spider.functions.clean_verifier"
            )(context["request"], self):
                _abilities.add("verify")
            if self.updateable_by.filter(
                id=context["request"].auth_token.referrer_id
            ).exists():
                _abilities.add("push_update")

        return _abilities

    def map_data(self, name, field, data, graph, context):
        if (
            field.__class__.__name__ == "AnchorField" and
            field.use_default_anchor
        ):
            if data is None:
                return URIRef(self.get_primary_anchor(graph, context))
        return super().map_data(name, field, data, graph, context)

    @csrf_exempt
    def access_verify(self, **kwargs):
        # full url to result
        verified = kwargs["request"].POST.get("url", "")
        if "://" not in verified:
            return HttpResponse("invalid url", status=400)
        if verified in self.verified_by:
            return self.access_view(**kwargs)
        self.verified_by.append(verified)
        self.clean()
        self.save()

        return self.access_view(**kwargs)

    def access_push_update(self, **kwargs):
        kwargs["legend"] = escape(
            _("Update \"%s\" (push)") % self.__str__()
        )
        kwargs["inner_form"] = False
        return self.access_update(**kwargs)

    def access(self, context):
        if context["scope"] not in {"add", "view"}:
            context["extra_outer_forms"] = ["request_verification_form"]
        return super().access(context)

    def get_form(self, scope):
        from .forms import SpiderTagForm
        if scope == "add":
            return SpiderTagForm
        else:
            return self.layout.get_form()

    def get_references(self):
        if not getattr(self, "layout", None):
            return super().get_references()
        if self._cached_references is None:
            form = self.layout.get_form()(
                initial=self.tagdata.copy(),
                instance=self,
                uc=self.associated.usercomponent
            )
            _cached_references, needs_update = form.calc_references(True)
            if needs_update:
                self.associated.save(
                    update_fields=[
                        "attached_to_primary_anchor"
                    ]
                )
            self._cached_references = _cached_references
        return chain(
            self._cached_references,
            super().get_references()
        )

    def get_form_kwargs(self, instance=None, **kwargs):
        if kwargs["scope"] == "add":
            ret = super().get_form_kwargs(
                instance=instance,
                **kwargs
            )
            ret["user"] = self.associated.usercomponent.user
        else:
            ret = super().get_form_kwargs(
                **kwargs
            )
            ret["initial"] = self.tagdata.copy()
            ret["request"] = kwargs["request"]
            ret["uc"] = self.associated.usercomponent
        return ret

    def encode_verifiers(self):
        return "".join(
            map(
                lambda x: "verified_by={}\x1e".format(x),
                self.verified_by
            )
        )

    def get_info(self):
        return "{}{}tag={}\x1e".format(
            super().get_info(unique=self.primary, unlisted=False),
            self.encode_verifiers(),
            self.layout.name
        )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self._cached_references = None



"""
Form messages
"""


__all__ = [
    "help_text_static_token", "help_text_features_components",
    "help_text_features_contents", "time_help_text", "travel_protection"
]

from django.utils.translation import gettext_lazy as _


help_text_static_token = _("""Generate a new static token with variable strength<br/>
Tokens protect against bruteforce and attackers<br/>
If you have problems with attackers (because they know the token),
you can invalidate it with this option and/or add protections<br/>
<table style="color:red;">
<tr>
<td style="vertical-align: top">Warning:</td><td>this removes also access for all services you gave the link<td/>
</tr>
<tr>
<td style="vertical-align: top">Warning:</td><td>public user components disclose static token, regenerate after removing
public attribute from component for restoring protection</td>
</tr>
</table>
""")  # noqa: E501


help_text_features_components = _("""
Features which should be active on component and user content<br/>
Note: persistent Features (=features which use a persistent token) require and enable "Persistence"
""")  # noqa: E501


help_text_features_contents = _("""
Features which should be active on user content<br/>
Note: persistent Features (=features which use a persistent token) require the active "Persistence" Feature on usercomponent. They are elsewise excluded.
""")  # noqa: E501


time_help_text = _(
    "Time in \"{timezone}\" timezone"
)


travel_protection = _(
    "Hide: Hide protected contents and components<br/>"
    "Hide if triggered: Hide protected contents and components if triggered<br/>"  # noqa: E501
    "Disable: Disable Login (not available on Self-Protection (useless))<br/>"
    "Disable if triggered: Hide protected contents and components and disable login if triggered<br/>"  # noqa: E501
    "Wipe: Wipe protected content on login (maybe not available)<br/>"
    "Wipe User: destroy user on login (maybe not available)<br/><br/>"
    "Note: login protections work only on login. Except hide and disable protections, which works also for logged in users<br/>"  # noqa: E501
    "Note: the hide and disable protections disable also admin access when active (elsewise easy circumventable)"  # noqa: E501
)
